@@ -478,7 +478,7 @@ static int net_tls_client(lua_State *L)
 
 	const char *full_addr = NULL;
 	bool pin_exists = false;
-	bool ca_file_exists = false;
+	bool hostname_exists = false;
 	if ((lua_gettop(L) == 1) && lua_isstring(L, 1)) {
 		full_addr = lua_tostring(L, 1);
 	} else if ((lua_gettop(L) == 2) && lua_isstring(L, 1) && lua_istable(L, 2)) {
@@ -486,12 +486,12 @@ static int net_tls_client(lua_State *L)
 		pin_exists = true;
 	} else if ((lua_gettop(L) == 3) && lua_isstring(L, 1) && lua_istable(L, 2)) {
 		full_addr = lua_tostring(L, 1);
-		ca_file_exists = true;
+		hostname_exists = true;
 	} else if ((lua_gettop(L) == 4) && lua_isstring(L, 1) &&
 		    lua_istable(L, 2) && lua_istable(L, 3)) {
 		full_addr = lua_tostring(L, 1);
 		pin_exists = true;
-		ca_file_exists = true;
+		hostname_exists = true;
 	} else {
 		format_error(L, "net.tls_client takes one parameter (\"address\"), two parameters (\"address\",\"pin\"), three parameters (\"address\", \"ca_file\", \"hostname\") or four ones: (\"address\", \"pin\", \"ca_file\", \"hostname\")");
 		lua_error(L);
@@ -508,9 +508,10 @@ static int net_tls_client(lua_State *L)
 		port = 853;
 	}
 
-	if (!pin_exists && !ca_file_exists) {
+	if (!pin_exists && !hostname_exists) {
 		int r = tls_client_params_set(&net->tls_client_params,
-					      addr, port, NULL, NULL, NULL);
+					      addr, port, NULL,
+					      TLS_CLIENT_PARAM_NONE);
 		if (r != 0) {
 			lua_pushstring(L, kr_strerror(r));
 			lua_error(L);
@@ -528,7 +529,8 @@ static int net_tls_client(lua_State *L)
 			/* pin now at index -1, key at index -2*/
 			const char *pin = lua_tostring(L, -1);
 			int r = tls_client_params_set(&net->tls_client_params,
-						      addr, port, NULL, NULL, pin);
+						      addr, port, pin,
+						      TLS_CLIENT_PARAM_PIN);
 			if (r != 0) {
 				lua_pushstring(L, kr_strerror(r));
 				lua_error(L);
@@ -539,7 +541,7 @@ static int net_tls_client(lua_State *L)
 
 	int ca_table_index = 2;
 	int hostname_table_index = 3;
-	if (ca_file_exists) {
+	if (hostname_exists) {
 		if (pin_exists) {
 			ca_table_index = 3;
 			hostname_table_index = 4;
@@ -549,12 +551,14 @@ static int net_tls_client(lua_State *L)
 		return 1;
 	}
 
-	/* iterate over ca filenames */
+	/* iterate over hostnames,
+	 * it must be done before iterating over ca filenames */
 	lua_pushnil(L);
-	while (lua_next(L, ca_table_index)) {
-		const char *ca_file = lua_tostring(L, -1);
+	while (lua_next(L, hostname_table_index)) {
+		const char *hostname = lua_tostring(L, -1);
 		int r = tls_client_params_set(&net->tls_client_params,
-					      addr, port, ca_file, NULL, NULL);
+					      addr, port, hostname,
+					      TLS_CLIENT_PARAM_HOSTNAME);
 		if (r != 0) {
 			lua_pushstring(L, kr_strerror(r));
 			lua_error(L);
@@ -563,20 +567,34 @@ static int net_tls_client(lua_State *L)
 		lua_pop(L, 1);
 	}
 
-	/* iterate over hostnames */
+	/* iterate over ca filenames */
 	lua_pushnil(L);
-	while (lua_next(L, hostname_table_index)) {
-		const char *hostname = lua_tostring(L, -1);
+	size_t num_of_ca_files = 0;
+	while (lua_next(L, ca_table_index)) {
+		const char *ca_file = lua_tostring(L, -1);
 		int r = tls_client_params_set(&net->tls_client_params,
-					      addr, port, NULL, hostname, NULL);
+					      addr, port, ca_file,
+					      TLS_CLIENT_PARAM_CA);
 		if (r != 0) {
 			lua_pushstring(L, kr_strerror(r));
 			lua_error(L);
 		}
+		num_of_ca_files += 1;
 		/* removes 'value'; keeps 'key' for next iteration */
 		lua_pop(L, 1);
 	}
 
+	if (num_of_ca_files == 0) {
+		/* No ca files were explicitly configured, so use system CA */
+		int r = tls_client_params_set(&net->tls_client_params,
+					      addr, port, NULL,
+					      TLS_CLIENT_PARAM_CA);
+		if (r != 0) {
+			lua_pushstring(L, kr_strerror(r));
+			lua_error(L);
+		}
+	}
+
 	lua_pushboolean(L, true);
 	return 1;
 }
@@ -624,6 +642,100 @@ static int net_tls_padding(lua_State *L)
 	return 1;
 }
 
+/** Shorter salt can't contain much entropy. */
+#define net_tls_sticket_MIN_SECRET_LEN 32
+
+static int net_tls_sticket_secret_string(lua_State *L)
+{
+	struct network *net = &engine_luaget(L)->net;
+
+	size_t secret_len;
+	const char *secret;
+
+	if (lua_gettop(L) == 0) {
+		/* Zero-length secret, implying random key. */
+		secret_len = 0;
+		secret = NULL;
+	} else {
+		if (lua_gettop(L) != 1 || !lua_isstring(L, 1)) {
+			lua_pushstring(L,
+				"net.tls_sticket_secret takes one parameter: (\"secret string\")");
+			lua_error(L);
+		}
+		secret = lua_tolstring(L, 1, &secret_len);
+		if (secret_len < net_tls_sticket_MIN_SECRET_LEN || !secret) {
+			lua_pushstring(L, "net.tls_sticket_secret - the secret is shorter than "
+						xstr(net_tls_sticket_MIN_SECRET_LEN) " bytes");
+			lua_error(L);
+		}
+	}
+
+	tls_session_ticket_ctx_destroy(net->tls_session_ticket_ctx);
+	net->tls_session_ticket_ctx =
+		tls_session_ticket_ctx_create(net->loop, secret, secret_len);
+	if (net->tls_session_ticket_ctx == NULL) {
+		lua_pushstring(L,
+			"net.tls_sticket_secret_string - can't create session ticket context");
+		lua_error(L);
+	}
+
+	lua_pushboolean(L, true);
+	return 1;
+}
+
+static int net_tls_sticket_secret_file(lua_State *L)
+{
+	if (lua_gettop(L) != 1 || !lua_isstring(L, 1)) {
+		lua_pushstring(L,
+			"net.tls_sticket_secret_file takes one parameter: (\"file name\")");
+		lua_error(L);
+	}
+
+	const char *file_name = lua_tostring(L, 1);
+	if (strlen(file_name) == 0) {
+		lua_pushstring(L, "net.tls_sticket_secret_file - empty file name");
+		lua_error(L);
+	}
+
+	FILE *fp = fopen(file_name, "r");
+	if (fp == NULL) {
+		lua_pushfstring(L, "net.tls_sticket_secret_file - can't open file '%s': %s",
+				file_name, strerror(errno));
+		lua_error(L);
+	}
+
+	char secret_buf[TLS_SESSION_TICKET_SECRET_MAX_LEN];
+	const size_t secret_len = fread(secret_buf, 1, sizeof(secret_buf), fp);
+	int err = ferror(fp);
+	if (err) {
+		lua_pushfstring(L,
+			"net.tls_sticket_secret_file - error reading from file '%s': %s",
+			file_name, strerror(err));
+		lua_error(L);
+	}
+	if (secret_len < net_tls_sticket_MIN_SECRET_LEN) {
+		lua_pushfstring(L,
+			"net.tls_sticket_secret_file - file '%s' is shorter than "
+				xstr(net_tls_sticket_MIN_SECRET_LEN) " bytes",
+			file_name);
+		lua_error(L);
+	}
+	fclose(fp);
+
+	struct network *net = &engine_luaget(L)->net;
+
+	tls_session_ticket_ctx_destroy(net->tls_session_ticket_ctx);
+	net->tls_session_ticket_ctx =
+		tls_session_ticket_ctx_create(net->loop, secret_buf, secret_len);
+	if (net->tls_session_ticket_ctx == NULL) {
+		lua_pushstring(L,
+			"net.tls_sticket_secret_file - can't create session ticket context");
+		lua_error(L);
+	}
+	lua_pushboolean(L, true);
+	return 1;
+}
+
 static int net_outgoing(lua_State *L, int family)
 {
 	struct worker_ctx *worker = wrk_luaget(L);
@@ -681,6 +793,37 @@ static int net_outgoing(lua_State *L, int family)
 static int net_outgoing_v4(lua_State *L) { return net_outgoing(L, AF_INET); }
 static int net_outgoing_v6(lua_State *L) { return net_outgoing(L, AF_INET6); }
 
+static int net_tcp_in_idle(lua_State *L)
+{
+	struct engine *engine = engine_luaget(L);
+	struct network *net = &engine->net;
+
+	/* Only return current idle timeout. */
+	if (lua_gettop(L) == 0) {
+		lua_pushnumber(L, net->tcp.in_idle_timeout);
+		return 1;
+	}
+
+	if ((lua_gettop(L) != 1)) {
+		lua_pushstring(L, "net.tcp_in_idle takes one parameter: (\"idle timeout\")");
+		lua_error(L);
+	}
+
+	if (lua_isnumber(L, 1)) {
+		int idle_timeout = lua_tointeger(L, 1);
+		if (idle_timeout <= 0) {
+			lua_pushstring(L, "net.tcp_in_idle parameter has to be positive number");
+			lua_error(L);
+		}
+		net->tcp.in_idle_timeout = idle_timeout;
+	} else {
+		lua_pushstring(L, "net.tcp_in_idle parameter has to be positive number");
+		lua_error(L);
+	}
+	lua_pushboolean(L, true);
+	return 1;
+}
+
 int lib_net(lua_State *L)
 {
 	static const luaL_Reg lib[] = {
@@ -694,14 +837,32 @@ int lib_net(lua_State *L)
 		{ "tls_server",   net_tls },
 		{ "tls_client",   net_tls_client },
 		{ "tls_padding",  net_tls_padding },
+		{ "tls_sticket_secret", net_tls_sticket_secret_string },
+		{ "tls_sticket_secret_file", net_tls_sticket_secret_file },
 		{ "outgoing_v4",  net_outgoing_v4 },
 		{ "outgoing_v6",  net_outgoing_v6 },
+		{ "tcp_in_idle",  net_tcp_in_idle },
 		{ NULL, NULL }
 	};
 	register_lib(L, "net", lib);
 	return 1;
 }
 
+
+
+/** @internal return cache, or throw lua error if not open */
+struct kr_cache * cache_assert_open(lua_State *L)
+{
+	struct engine *engine = engine_luaget(L);
+	struct kr_cache *cache = &engine->resolver.cache;
+	assert(cache);
+	if (!cache || !kr_cache_is_open(cache)) {
+		format_error(L, "no cache is open yet, use cache.open() or cache.size, etc.");
+		lua_error(L);
+	}
+	return cache;
+}
+
 /** Return available cached backends. */
 static int cache_backends(lua_State *L)
 {
@@ -719,12 +880,10 @@ static int cache_backends(lua_State *L)
 /** Return number of cached records. */
 static int cache_count(lua_State *L)
 {
-	struct engine *engine = engine_luaget(L);
-	const struct kr_cdb_api *api = engine->resolver.cache.api;
+	struct kr_cache *cache = cache_assert_open(L);
 
-	struct kr_cache *cache = &engine->resolver.cache;
-	int count = api->count(cache->db);
-	if (kr_cache_is_open(cache) && count >= 0) {
+	int count = cache->api->count(cache->db);
+	if (count >= 0) {
 		/* First key is a version counter, omit it if nonempty. */
 		lua_pushinteger(L, count ? count - 1 : 0);
 		return 1;
@@ -735,8 +894,7 @@ static int cache_count(lua_State *L)
 /** Return time of last checkpoint, or re-set it if passed `true`. */
 static int cache_checkpoint(lua_State *L)
 {
-	struct engine *engine = engine_luaget(L);
-	struct kr_cache *cache = &engine->resolver.cache;
+	struct kr_cache *cache = cache_assert_open(L);
 
 	if (lua_gettop(L) == 0) { /* Return the current value. */
 		lua_newtable(L);
@@ -762,8 +920,7 @@ static int cache_checkpoint(lua_State *L)
 /** Return cache statistics. */
 static int cache_stats(lua_State *L)
 {
-	struct engine *engine = engine_luaget(L);
-	struct kr_cache *cache = &engine->resolver.cache;
+	struct kr_cache *cache = cache_assert_open(L);
 	lua_newtable(L);
 	lua_pushnumber(L, cache->stats.hit);
 	lua_setfield(L, -2, "hit");
@@ -797,8 +954,7 @@ static const struct kr_cdb_api *cache_select(struct engine *engine, const char *
 
 static int cache_max_ttl(lua_State *L)
 {
-	struct engine *engine = engine_luaget(L);
-	struct kr_cache *cache = &engine->resolver.cache;
+	struct kr_cache *cache = cache_assert_open(L);
 
 	int n = lua_gettop(L);
 	if (n > 0) {
@@ -808,7 +964,7 @@ static int cache_max_ttl(lua_State *L)
 		}
 		uint32_t min = cache->ttl_min;
 		int64_t ttl = lua_tonumber(L, 1);
-		if (ttl < 0 || ttl <= min || ttl > UINT32_MAX) {
+		if (ttl < 0 || ttl < min || ttl > UINT32_MAX) {
 			format_error(L, "max_ttl must be larger than minimum TTL, and in range <1, " xstr(UINT32_MAX) ">'");
 			lua_error(L);
 		}
@@ -821,8 +977,7 @@ static int cache_max_ttl(lua_State *L)
 
 static int cache_min_ttl(lua_State *L)
 {
-	struct engine *engine = engine_luaget(L);
-	struct kr_cache *cache = &engine->resolver.cache;
+	struct kr_cache *cache = cache_assert_open(L);
 
 	int n = lua_gettop(L);
 	if (n > 0) {
@@ -832,7 +987,7 @@ static int cache_min_ttl(lua_State *L)
 		}
 		uint32_t max = cache->ttl_max;
 		int64_t ttl = lua_tonumber(L, 1);
-		if (ttl < 0 || ttl >= max || ttl > UINT32_MAX) {
+		if (ttl < 0 || ttl > max || ttl > UINT32_MAX) {
 			format_error(L, "min_ttl must be smaller than maximum TTL, and in range <0, " xstr(UINT32_MAX) ">'");
 			lua_error(L);
 		}
@@ -922,69 +1077,23 @@ static int cache_close(lua_State *L)
 
 #if 0
 /** @internal Prefix walk. */
-static int cache_prefixed(struct kr_cache *cache, const char *args, knot_db_val_t *results, int maxresults)
+static int cache_prefixed(struct kr_cache *cache, const char *prefix, bool exact_name,
+			  knot_db_val_t keyval[][2], int maxcount)
 {
-	/* Decode parameters */
-	uint8_t namespace = 'R';
-	char *extra = strchr(args, ' ');
-	if (extra != NULL) {
-		extra[0] = '\0';
-		namespace = extra[1];
-	}
-
 	/* Convert to domain name */
 	uint8_t buf[KNOT_DNAME_MAXLEN];
-	if (!knot_dname_from_str(buf, args, sizeof(buf))) {
+	if (!knot_dname_from_str(buf, prefix, sizeof(buf))) {
 		return kr_error(EINVAL);
 	}
-
 	/* Start prefix search */
-	int ret = kr_cache_match(cache, namespace, buf, results, maxresults);
-	kr_cache_sync(cache);
-	return ret;
-}
-
-/** @internal Delete iterated key. */
-static int cache_remove_prefix(struct kr_cache *cache, const char *args)
-{
-	/* Check if we can remove */
-	if (!cache || !cache->api || !cache->api->remove) {
-		return kr_error(ENOSYS);
-	}
-	static knot_db_val_t result_set[1000];
-	int ret = cache_prefixed(cache, args, result_set, 1000);
-	if (ret < 0) {
-		return ret;
-	}
-	/* Duplicate result set as we're going to remove it
-	 * which will invalidate result set. */
-	for (int i = 0; i < ret; ++i) {
-		void *dst = malloc(result_set[i].len);
-		if (!dst) {
-			return kr_error(ENOMEM);
-		}
-		memcpy(dst, result_set[i].data, result_set[i].len);
-		result_set[i].data = dst;
-	}
-	cache->api->remove(cache->db, result_set, ret);
-	kr_cache_sync(cache);
-	/* Free keys */
-	for (int i = 0; i < ret; ++i) {
-		free(result_set[i].data);
-	}
-	return ret;
+	return kr_cache_match(cache, buf, exact_name, keyval, maxcount);
 }
 #endif
 
 /** Prune expired/invalid records. */
 static int cache_prune(lua_State *L)
 {
-	struct engine *engine = engine_luaget(L);
-	struct kr_cache *cache = &engine->resolver.cache;
-	if (!kr_cache_is_open(cache)) {
-		return 0;
-	}
-
+	struct kr_cache *cache = cache_assert_open(L);
 	/* Check parameters */
 	int prune_max = UINT16_MAX;
 	int n = lua_gettop(L);
@@ -1006,34 +1115,12 @@ static int cache_prune(lua_State *L)
 	return 1;
 }
 
-/** Clear all records. */
-static int cache_clear(lua_State *L)
+/** Clear everything. */
+static int cache_clear_everything(lua_State *L)
 {
-	struct engine *engine = engine_luaget(L);
-	struct kr_cache *cache = &engine->resolver.cache;
-	if (!kr_cache_is_open(cache)) {
-		return 0;
-	}
+	struct kr_cache *cache = cache_assert_open(L);
 
-	/* Check parameters */
-	const char *args = NULL;
-	int n = lua_gettop(L);
-	if (n >= 1 && lua_isstring(L, 1)) {
-		args = lua_tostring(L, 1);
-	}
-
-	/* Clear a sub-tree in cache. */
-	if (args && strlen(args) > 0) {
-		int ret = kr_error(ENOSYS); // FIXME cache_remove_prefix(cache, args);
-		if (ret < 0) {
-			format_error(L, kr_strerror(ret));
-			lua_error(L);
-		}
-		lua_pushinteger(L, ret);
-		return 1;
-	}
-
-	/* Clear cache. */
+	/* Clear records and packets. */
 	int ret = kr_cache_clear(cache);
 	if (ret < 0) {
 		format_error(L, kr_strerror(ret));
@@ -1041,6 +1128,7 @@ static int cache_clear(lua_State *L)
 	}
 
 	/* Clear reputation tables */
+	struct engine *engine = engine_luaget(L);
 	lru_reset(engine->resolver.cache_rtt);
 	lru_reset(engine->resolver.cache_rep);
 	lru_reset(engine->resolver.cache_cookie);
@@ -1048,52 +1136,42 @@ static int cache_clear(lua_State *L)
 	return 1;
 }
 
+#if 0
 /** @internal Dump cache key into table on Lua stack. */
-static void cache_dump_key(lua_State *L, knot_db_val_t *key)
+static void cache_dump(lua_State *L, knot_db_val_t keyval[])
 {
-	char buf[KNOT_DNAME_MAXLEN];
-	/* Extract type */
-	uint16_t type = 0;
-	const char *endp = (const char *)key->data + key->len - sizeof(uint16_t);
-	memcpy(&type, endp, sizeof(uint16_t));
-	endp -= 1;
-	/* Extract domain name */
-	char *dst = buf;
-	const char *scan = endp - 1;
-	while (scan > (const char *)key->data) {
-		if (*scan == '\0') {
-			const size_t lblen = endp - scan - 1;
-			memcpy(dst, scan + 1, lblen);
-			dst += lblen;
-			*dst++ = '.';
-			endp = scan;
-		}
-		--scan;
+	knot_dname_t dname[KNOT_DNAME_MAXLEN];
+	char name[KNOT_DNAME_TXT_MAXLEN];
+	uint16_t type;
+
+	int ret = kr_unpack_cache_key(keyval[0], dname, &type);
+	if (ret < 0) {
+		return;
 	}
-	memcpy(dst, scan + 1, endp - scan);
+
+	ret = !knot_dname_to_str(name, dname, sizeof(name));
+	assert(!ret);
+	if (ret) return;
+
 	/* If name typemap doesn't exist yet, create it */
-	lua_getfield(L, -1, buf);
+	lua_getfield(L, -1, name);
 	if (lua_isnil(L, -1)) {
 		lua_pop(L, 1);
 		lua_newtable(L);
 	}
 	/* Append to typemap */
-	char type_buf[16] = { '\0' };
+	char type_buf[KR_RRTYPE_STR_MAXLEN] = { '\0' };
 	knot_rrtype_to_string(type, type_buf, sizeof(type_buf));
 	lua_pushboolean(L, true);
 	lua_setfield(L, -2, type_buf);
 	/* Set name typemap */
-	lua_setfield(L, -2, buf);
+	lua_setfield(L, -2, name);
 }
 
-/** Query cached records. */
+/** Query cached records.  TODO: fix caveats in ./README.rst documentation? */
 static int cache_get(lua_State *L)
 {
-	struct engine *engine = engine_luaget(L);
-	struct kr_cache *cache = &engine->resolver.cache;
-	if (!kr_cache_is_open(cache)) {
-		return 0;
-	}
+	//struct kr_cache *cache = cache_assert_open(L); // to be fixed soon
 
 	/* Check parameters */
 	int n = lua_gettop(L);
@@ -1103,9 +1181,9 @@ static int cache_get(lua_State *L)
 	}
 
 	/* Retrieve set of keys */
-	//const char *args = lua_tostring(L, 1);
-	static knot_db_val_t result_set[100];
-	int ret = kr_error(ENOSYS); // FIXME cache_prefixed(cache, args, result_set, 100);
+	const char *prefix = lua_tostring(L, 1);
+	knot_db_val_t keyval[100][2];
+	int ret = cache_prefixed(cache, prefix, false/*FIXME*/, keyval, 100);
 	if (ret < 0) {
 		format_error(L, kr_strerror(ret));
 		lua_error(L);
@@ -1113,10 +1191,18 @@ static int cache_get(lua_State *L)
 	/* Format output */
 	lua_newtable(L);
 	for (int i = 0; i < ret; ++i) {
-		cache_dump_key(L, &result_set[i]);
+		cache_dump(L, keyval[i]);
 	}
 	return 1;
 }
+#endif
+static int cache_get(lua_State *L)
+{
+	int ret = kr_error(ENOSYS);
+	format_error(L, kr_strerror(ret));
+	lua_error(L);
+	return ret;
+}
 
 /** Set time interval for cleaning rtt cache.
  * Servers with score >= KR_NS_TIMEOUTED will be cleaned after
@@ -1179,16 +1265,7 @@ static int cache_zone_import(lua_State *L)
 		goto finish;
 	}
 
-	struct engine *engine = engine_luaget(L);
-	if (!engine) {
-		strncpy(msg, "internal error, empty engine pointer", sizeof(msg));
-		goto finish;
-	}
-	struct kr_cache *cache = &engine->resolver.cache;
-	if (!kr_cache_is_open(cache)) {
-		strncpy(msg, "cache isn't open", sizeof(msg));
-		goto finish;
-	}
+	(void)cache_assert_open(L); /* just check it in advance */
 
 	/* Check parameters */
 	int n = lua_gettop(L);
@@ -1245,8 +1322,8 @@ int lib_cache(lua_State *L)
 		{ "open",   cache_open },
 		{ "close",  cache_close },
 		{ "prune",  cache_prune },
-		{ "clear",  cache_clear },
-		{ "get",    cache_get },
+		{ "clear_everything", cache_clear_everything },
+		{ "get",     cache_get },
 		{ "max_ttl", cache_max_ttl },
 		{ "min_ttl", cache_min_ttl },
 		{ "ns_tout", cache_ns_tout },
@@ -1540,11 +1617,12 @@ static int wrk_resolve(lua_State *L)
 	}
 	knot_pkt_put_question(pkt, dname, rrclass, rrtype);
 	knot_wire_set_rd(pkt->wire);
+	knot_wire_set_ad(pkt->wire);
 
 	/* Add OPT RR */
 	pkt->opt_rr = knot_rrset_copy(worker->engine->resolver.opt_rr, NULL);
 	if (!pkt->opt_rr) {
-		knot_pkt_free(&pkt);
+		knot_pkt_free(pkt);
 		return kr_error(ENOMEM);
 	}
 	if (options->DNSSEC_WANT) {
@@ -1558,8 +1636,8 @@ static int wrk_resolve(lua_State *L)
 	/* Create task and start with a first question */
 	struct qr_task *task = worker_resolve_start(worker, pkt, *options);
 	if (!task) {
-		knot_rrset_free(&pkt->opt_rr, NULL);
-		knot_pkt_free(&pkt);
+		knot_rrset_free(pkt->opt_rr, NULL);
+		knot_pkt_free(pkt);
 		lua_pushstring(L, "couldn't create a resolution request");
 		lua_error(L);
 	}
@@ -1574,8 +1652,8 @@ static int wrk_resolve(lua_State *L)
 	/* Start execution */
 	int ret = worker_resolve_exec(task, pkt);
 	lua_pushboolean(L, ret == 0);
-	knot_rrset_free(&pkt->opt_rr, NULL);
-	knot_pkt_free(&pkt);
+	knot_rrset_free(pkt->opt_rr, NULL);
+	knot_pkt_free(pkt);
 	return 1;
 }
 
