@@ -19,21 +19,21 @@
  * this program.  If not, see <http://www.gnu.org/licenses/>.
  */
 
-#include <gnutls/gnutls.h>
-#include <gnutls/x509.h>
 #include <gnutls/abstract.h>
 #include <gnutls/crypto.h>
-#include <stdlib.h>
-#include <errno.h>
-#include <assert.h>
 #include <gnutls/gnutls.h>
+#include <gnutls/x509.h>
 #include <uv.h>
 
-#include <contrib/ucw/lib.h>
+#include <assert.h>
+#include <errno.h>
+#include <stdlib.h>
+
+#include "contrib/ucw/lib.h"
 #include "contrib/base64.h"
-#include "daemon/worker.h"
-#include "daemon/tls.h"
 #include "daemon/io.h"
+#include "daemon/tls.h"
+#include "daemon/worker.h"
 
 #define EPHEMERAL_CERT_EXPIRATION_SECONDS_RENEW_BEFORE 60*60*24*7
 #define GNUTLS_PIN_MIN_VERSION  0x030400
@@ -59,7 +59,9 @@ static int kres_gnutls_set_priority(gnutls_session_t session) {
 	static const char * const priorities =
 		"NORMAL:" /* GnuTLS defaults */
 		"-VERS-TLS1.0:-VERS-TLS1.1:" /* TLS 1.2 and higher */
-		"-COMP-ALL:+COMP-NULL"; /* no compression*/
+		 /* Some distros by default allow features that are considered
+		  * too insecure nowadays, so let's disable them explicitly. */
+		"-VERS-SSL3.0:-ARCFOUR-128:-COMP-ALL:+COMP-NULL";
 	const char *errpos = NULL;
 	int err = gnutls_priority_set_direct(session, priorities, &errpos);
 	if (err != GNUTLS_E_SUCCESS) {
@@ -158,6 +160,12 @@ struct tls_ctx_t *tls_new(struct worker_ctx *worker)
 	gnutls_transport_set_pull_function(tls->c.tls_session, kres_gnutls_pull);
 	gnutls_transport_set_push_function(tls->c.tls_session, worker_gnutls_push);
 	gnutls_transport_set_ptr(tls->c.tls_session, tls);
+
+	if (net->tls_session_ticket_ctx) {
+		tls_session_ticket_enable(net->tls_session_ticket_ctx,
+					  tls->c.tls_session);
+	}
+
 	return tls;
 }
 
@@ -299,11 +307,22 @@ int tls_process(struct worker_ctx *worker, uv_stream_t *handle, const uint8_t *b
 	}
 
 	int submitted = 0;
+	bool is_retrying = false;
+	uint64_t retrying_start = 0;
 	while (true) {
 		ssize_t count = gnutls_record_recv(tls_p->tls_session, tls_p->recv_buf, sizeof(tls_p->recv_buf));
 		if (count == GNUTLS_E_AGAIN) {
 			break;    /* No data available */
-		} else if (count == GNUTLS_E_INTERRUPTED) {
+		} else if (count == GNUTLS_E_INTERRUPTED ||
+			   count == GNUTLS_E_REHANDSHAKE) {
+			if (!is_retrying) {
+				is_retrying = true;
+				retrying_start = kr_now();
+			}
+			uint64_t elapsed = kr_now() - retrying_start;
+			if (elapsed > TLS_MAX_HANDSHAKE_TIME) {
+				return kr_error(EIO);
+			}
 			continue; /* Try reading again */
 		} else if (count < 0) {
 			kr_log_verbose("[%s] gnutls_record_recv failed: %s (%zd)\n",
@@ -576,6 +595,10 @@ static int client_paramlist_entry_clear(const char *k, void *v, void *baton)
 		gnutls_certificate_free_credentials(entry->credentials);
 	}
 
+	if (entry->session_data.data) {
+		gnutls_free(entry->session_data.data);
+	}
+
 	free(entry);
 
 	return 0;
@@ -583,12 +606,22 @@ static int client_paramlist_entry_clear(const char *k, void *v, void *baton)
 
 int tls_client_params_set(map_t *tls_client_paramlist,
 			  const char *addr, uint16_t port,
-			  const char *ca_file, const char *hostname, const char *pin)
+			  const char *param, tls_client_param_t param_type)
 {
 	if (!tls_client_paramlist || !addr) {
 		return kr_error(EINVAL);
 	}
 
+	/* TLS_CLIENT_PARAM_CA can be empty */
+	if (param_type == TLS_CLIENT_PARAM_HOSTNAME ||
+	    param_type == TLS_CLIENT_PARAM_PIN) {
+		if (param == NULL || param[0] == 0) {
+			return kr_error(EINVAL);
+		}
+	}
+
+	/* Parameters are OK */
+
 	char key[INET6_ADDRSTRLEN + 6];
 	size_t keylen = sizeof(key);
 	if (kr_straddr_join(addr, port, key, &keylen) != kr_ok()) {
@@ -615,40 +648,9 @@ int tls_client_params_set(map_t *tls_client_paramlist,
 	}
 
 	int ret = kr_ok();
-	if (ca_file && ca_file[0] != 0) {
-		bool already_exists = false;
-		for (size_t i = 0; i < entry->ca_files.len; ++i) {
-			if (strcmp(entry->ca_files.at[i], ca_file) == 0) {
-				kr_log_error("[tls_client] error: ca file '%s'for address '%s' already was set, ignoring\n", ca_file, key);
-				already_exists = true;
-				break;
-			}
-		}
-		if (!already_exists) {
-			const char *value = strdup(ca_file);
-			if (!value) {
-				ret = kr_error(ENOMEM);
-			} else if (array_push(entry->ca_files, value) < 0) {
-				free ((void *)value);
-				ret = kr_error(ENOMEM);
-			} else {
-				int res = gnutls_certificate_set_x509_trust_file(entry->credentials, value,
-										 GNUTLS_X509_FMT_PEM);
-				if (res <= 0) {
-					kr_log_error("[tls_client] failed to import certificate file '%s' (%s)\n",
-						     value, gnutls_strerror_name(res));
-					/* value will be freed at cleanup */
-					ret = kr_error(EINVAL);
-				} else {
-					kr_log_verbose("[tls_client] imported %d certs from file '%s'\n",
-							res, value);
 
-				}
-			}
-		}
-	}
-
-	if ((ret == kr_ok()) && hostname && hostname[0] != 0) {
+	if (param_type == TLS_CLIENT_PARAM_HOSTNAME) {
+		const char *hostname = param;
 		bool already_exists = false;
 		for (size_t i = 0; i < entry->hostnames.len; ++i) {
 			if (strcmp(entry->hostnames.at[i], hostname) == 0) {
@@ -666,9 +668,59 @@ int tls_client_params_set(map_t *tls_client_paramlist,
 				ret = kr_error(ENOMEM);
 			}
 		}
-	}
+	} else if (param_type == TLS_CLIENT_PARAM_CA) {
+		/* Import ca files only when hostname is already set */
+		if (entry->hostnames.len == 0) {
+			return kr_error(ENOENT);
+		}
+		const char *ca_file = param;
+		bool already_exists = false;
+		for (size_t i = 0; i < entry->ca_files.len; ++i) {
+			const char *imported_ca = entry->ca_files.at[i];
+			if (imported_ca[0] == 0 && (ca_file == NULL || ca_file[0] == 0)) {
+				kr_log_error("[tls_client] error: system ca for address '%s' already was set, ignoring\n", key);
+				already_exists = true;
+				break;
+			} else if (strcmp(imported_ca, ca_file) == 0) {
+				kr_log_error("[tls_client] error: ca file '%s' for address '%s' already was set, ignoring\n", ca_file, key);
+				already_exists = true;
+				break;
+			}
+		}
+		if (!already_exists) {
+			const char *value = strdup(ca_file != NULL ? ca_file : "");
+			if (!value) {
+				ret = kr_error(ENOMEM);
+			} else if (array_push(entry->ca_files, value) < 0) {
+				free ((void *)value);
+				ret = kr_error(ENOMEM);
+			} else if (value[0] == 0) {
+				int res = gnutls_certificate_set_x509_system_trust (entry->credentials);
+				if (res <= 0) {
+					kr_log_error("[tls_client] failed to import certs from system store (%s)\n",
+						     gnutls_strerror_name(res));
+					/* value will be freed at cleanup */
+					ret = kr_error(EINVAL);
+				} else {
+					kr_log_verbose("[tls_client] imported %d certs from system store\n", res);
+				}
+			} else {
+				int res = gnutls_certificate_set_x509_trust_file(entry->credentials, value,
+										 GNUTLS_X509_FMT_PEM);
+				if (res <= 0) {
+					kr_log_error("[tls_client] failed to import certificate file '%s' (%s)\n",
+						     value, gnutls_strerror_name(res));
+					/* value will be freed at cleanup */
+					ret = kr_error(EINVAL);
+				} else {
+					kr_log_verbose("[tls_client] imported %d certs from file '%s'\n",
+							res, value);
 
-	if ((ret == kr_ok()) && pin && pin[0] != 0) {
+				}
+			}
+		}
+	} else if (param_type == TLS_CLIENT_PARAM_PIN) {
+		const char *pin = param;
 		for (size_t i = 0; i < entry->pins.len; ++i) {
 			if (strcmp(entry->pins.at[i], pin) == 0) {
 				kr_log_error("[tls_client] warning: pin '%s' for address '%s' already was set, ignoring\n", pin, key);
@@ -896,6 +948,12 @@ int tls_client_connect_start(struct tls_client_ctx_t *client_ctx,
 	ctx->handshake_state = TLS_HS_IN_PROGRESS;
 	ctx->session = session;
 
+	struct tls_client_paramlist_entry *tls_params = client_ctx->params;
+	if (tls_params->session_data.data != NULL) {
+		gnutls_session_set_data(ctx->tls_session, tls_params->session_data.data,
+					tls_params->session_data.size);
+	}
+
 	int ret = gnutls_handshake(ctx->tls_session);
 	if (ret == GNUTLS_E_SUCCESS) {
 		return kr_ok();
@@ -921,7 +979,7 @@ int tls_set_hs_state(struct tls_common_ctx *ctx, tls_hs_state_t state)
 }
 
 int tls_client_ctx_set_params(struct tls_client_ctx_t *ctx,
-			      const struct tls_client_paramlist_entry *entry,
+			      struct tls_client_paramlist_entry *entry,
 			      struct session *session)
 {
 	if (!ctx) {
