@@ -531,6 +531,7 @@ static struct request_ctx *request_create(struct worker_ctx *worker,
 
 	struct kr_request *req = &ctx->req;
 	req->pool = pool;
+	req->vars_ref = LUA_NOREF;
 
 	/* Remember query source addr */
 	if (!addr || (addr->sa_family != AF_INET && addr->sa_family != AF_INET6)) {
@@ -614,6 +615,20 @@ static int request_start(struct request_ctx *ctx, knot_pkt_t *query)
 static void request_free(struct request_ctx *ctx)
 {
 	struct worker_ctx *worker = ctx->worker;
+	/* Dereference any Lua vars table if exists */
+	if (ctx->req.vars_ref != LUA_NOREF) {
+		lua_State *L = worker->engine->L;
+		/* Get worker variables table */
+		lua_rawgeti(L, LUA_REGISTRYINDEX, worker->vars_table_ref);
+		/* Get next free element (position 0) and store it under current reference (forming a list) */
+		lua_rawgeti(L, -1, 0);
+		lua_rawseti(L, -2, ctx->req.vars_ref);
+		/* Set current reference as the next free element */
+		lua_pushinteger(L, ctx->req.vars_ref);
+		lua_rawseti(L, -2, 0);
+		lua_pop(L, 1);
+		ctx->req.vars_ref = LUA_NOREF;
+	}
 	/* Return mempool to ring or free it if it's full */
 	pool_release(worker, ctx->req.pool.ctx);
 	/* @note The 'task' is invalidated from now on. */
@@ -1082,15 +1097,38 @@ static int session_tls_hs_cb(struct session *session, int status)
 	struct worker_ctx *worker = get_worker();
 	union inaddr *peer = &session->peer;
 	int deletion_res = worker_del_tcp_waiting(worker, &peer->ip);
+	int ret = kr_ok();
 
 	if (status) {
 		kr_nsrep_update_rtt(NULL, &peer->ip, KR_NS_DEAD,
 				    worker->engine->resolver.cache_rtt,
 				    KR_NS_UPDATE_NORESET);
-		return kr_ok();
+		return ret;
 	}
 
-	int ret = worker_add_tcp_connected(worker, &peer->ip, session);
+	/* handshake was completed successfully */
+	struct tls_client_ctx_t *tls_client_ctx = session->tls_client_ctx;
+	struct tls_client_paramlist_entry *tls_params = tls_client_ctx->params;
+	gnutls_session_t tls_session = tls_client_ctx->c.tls_session;
+	if (gnutls_session_is_resumed(tls_session) != 0) {
+		kr_log_verbose("[tls_client] TLS session has resumed\n");
+	} else {
+		kr_log_verbose("[tls_client] TLS session has not resumed\n");
+		/* session wasn't resumed, delete old session data ... */
+		if (tls_params->session_data.data != NULL) {
+			gnutls_free(tls_params->session_data.data);
+			tls_params->session_data.data = NULL;
+			tls_params->session_data.size = 0;
+		}
+		/* ... and get the new session data */
+		gnutls_datum_t tls_session_data = { NULL, 0 };
+		ret = gnutls_session_get_data2(tls_session, &tls_session_data);
+		if (ret == 0) {
+			tls_params->session_data = tls_session_data;
+		}
+	}
+
+	ret = worker_add_tcp_connected(worker, &peer->ip, session);
 	if (deletion_res == kr_ok() && ret == kr_ok()) {
 		ret = session_next_waiting_send(session);
 	} else {
@@ -1269,8 +1307,7 @@ static void on_tcp_connect_timeout(uv_timer_t *timer)
 
 	while (session->waiting.len > 0) {
 		struct qr_task *task = session->waiting.at[0];
-		struct request_ctx *ctx = task->ctx;
-		assert(ctx);
+		assert(task->ctx);
 		task->timeouts += 1;
 		worker->stats.timeout += 1;
 		session_del_tasks(session, task);
@@ -1329,9 +1366,7 @@ static void on_tcp_watchdog_timeout(uv_timer_t *timer)
 static void on_udp_timeout(uv_timer_t *timer)
 {
 	struct session *session = timer->data;
-
-	uv_handle_t *handle = session->handle;
-	assert(handle->data == session);
+	assert(session->handle->data == session);
 
 	uv_timer_stop(timer);
 	assert(session->tasks.len == 1);
@@ -1384,8 +1419,14 @@ static uv_handle_t *retransmit(struct qr_task *task)
 			return ret;
 		}
 		ret = ioreq_spawn(task, SOCK_DGRAM, choice->sin6_family);
-		if (ret &&
-		    qr_task_send(task, ret, (struct sockaddr *)choice,
+		if (!ret) {
+			return ret;
+		}
+		struct sockaddr *addr = (struct sockaddr *)choice;
+		struct session *session = ret->data;
+		assert (session->peer.ip.sa_family == AF_UNSPEC && session->outgoing);
+		memcpy(&session->peer, addr, sizeof(session->peer));
+		if (qr_task_send(task, ret, (struct sockaddr *)choice,
 				 task->pktbuf) == 0) {
 			task->addrlist_turn = (task->addrlist_turn + 1) %
 					      task->addrlist_count; /* Round robin */
@@ -1435,7 +1476,7 @@ static void subreq_finalize(struct qr_task *task, const struct sockaddr *packet_
 	if (klen > 0) {
 		void *val_deleted;
 		int ret = trie_del(task->ctx->worker->subreq_out, key, klen, &val_deleted);
-		assert(ret == KNOT_EOK && val_deleted == task);
+		assert(ret == KNOT_EOK && val_deleted == task); (void)ret;
 	}
 	/* Notify waiting tasks. */
 	struct kr_query *leader_qry = array_tail(task->ctx->req.rplan.pending);
@@ -1520,6 +1561,18 @@ static int qr_task_finalize(struct qr_task *task, int state)
 			        ctx->req.answer);
 	if (res != kr_ok()) {
 		(void) qr_task_on_send(task, NULL, kr_error(EIO));
+		/* Since source session is erroneous detach all tasks. */
+		while (source_session->tasks.len > 0) {
+			struct qr_task *t = source_session->tasks.at[0];
+			struct request_ctx *c = t->ctx;
+			assert(c->source.session == source_session);
+			c->source.session = NULL;
+			/* Don't finalize them as there can be other tasks
+			 * waiting for answer to this particular task.
+			 * (ie. task->leading is true) */
+			session_del_tasks(source_session, t);
+		}
+		session_close(source_session);
 	} else if (handle->type == UV_TCP) {
 		/* Don't try to close source session at least
 		 * retry_interval_for_timeout_timer milliseconds */
@@ -1816,16 +1869,17 @@ static int parse_packet(knot_pkt_t *query)
 
 	/* Parse query packet. */
 	int ret = knot_pkt_parse(query, 0);
-	if (ret != KNOT_EOK) {
-		return kr_error(EPROTO); /* Ignore malformed query. */
+	if (ret == KNOT_ETRAIL) {
+		/* Extra data after message end. */
+		ret = kr_error(EMSGSIZE);
+	} else if (ret != KNOT_EOK) {
+		/* Malformed query. */
+		ret = kr_error(EPROTO);
+	} else {
+		ret = kr_ok();
 	}
 
-	/* Check if at least header is parsed. */
-	if (query->parsed < query->size) {
-		return kr_error(EMSGSIZE);
-	}
-
-	return kr_ok();
+	return ret;
 }
 
 static struct qr_task* find_task(const struct session *session, uint16_t msg_id)
@@ -1862,7 +1916,7 @@ int worker_submit(struct worker_ctx *worker, uv_handle_t *handle,
 	 * or resume if this is subrequest */
 	struct qr_task *task = NULL;
 	if (!session->outgoing) { /* request from a client */
-		/* Ignore badly formed queries or responses. */
+		/* Ignore badly formed queries. */
 		if (!query || ret != 0 || knot_wire_get_qr(query->wire)) {
 			if (query) worker->stats.dropped += 1;
 			return kr_error(EILSEQ);
@@ -1885,6 +1939,11 @@ int worker_submit(struct worker_ctx *worker, uv_handle_t *handle,
 		}
 		addr = NULL;
 	} else if (query) { /* response from upstream */
+		if ((ret != kr_ok() && ret != kr_error(EMSGSIZE)) ||
+		    !knot_wire_get_qr(query->wire)) {
+			/* Ignore badly formed responses. */
+			return kr_error(EILSEQ);
+		}
 		task = find_task(session, knot_wire_get_id(query->wire));
 		if (task == NULL) {
 			return kr_error(ENOENT);
@@ -1931,21 +1990,20 @@ static int worker_add_tcp_connected(struct worker_ctx *worker,
 				    const struct sockaddr* addr,
 				    struct session *session)
 {
+#ifndef NDEBUG
 	assert(addr);
 	const char *key = tcpsess_key(addr);
 	assert(key);
 	assert(map_contains(&worker->tcp_connected, key) == 0);
+#endif
 	return map_add_tcp_session(&worker->tcp_connected, addr, session);
 }
 
 static int worker_del_tcp_connected(struct worker_ctx *worker,
 				    const struct sockaddr* addr)
 {
-	assert(addr);
-	const char *key = tcpsess_key(addr);
-	assert(key);
-	int ret = map_del_tcp_session(&worker->tcp_connected, addr);
-	return ret;
+	assert(addr && tcpsess_key(addr));
+	return map_del_tcp_session(&worker->tcp_connected, addr);
 }
 
 static struct session* worker_find_tcp_connected(struct worker_ctx *worker,
@@ -1958,22 +2016,20 @@ static int worker_add_tcp_waiting(struct worker_ctx *worker,
 				  const struct sockaddr* addr,
 				  struct session *session)
 {
+#ifndef NDEBUG
 	assert(addr);
 	const char *key = tcpsess_key(addr);
 	assert(key);
 	assert(map_contains(&worker->tcp_waiting, key) == 0);
-	int ret = map_add_tcp_session(&worker->tcp_waiting, addr, session);
-	return ret;
+#endif
+	return map_add_tcp_session(&worker->tcp_waiting, addr, session);
 }
 
 static int worker_del_tcp_waiting(struct worker_ctx *worker,
 				  const struct sockaddr* addr)
 {
-	assert(addr);
-	const char *key = tcpsess_key(addr);
-	assert(key);
-	int ret = map_del_tcp_session(&worker->tcp_waiting, addr);
-	return ret;
+	assert(addr && tcpsess_key(addr));
+	return map_del_tcp_session(&worker->tcp_waiting, addr);
 }
 
 static struct session* worker_find_tcp_waiting(struct worker_ctx *worker,
@@ -2290,13 +2346,16 @@ int worker_process_tcp(struct worker_ctx *worker, uv_stream_t *handle,
 		/* Message was assembled, clear temporary. */
 		session->buffering = NULL;
 		session->msg_hdr_idx = 0;
+		const struct sockaddr *addr = NULL;
+		knot_pkt_t *pkt = pkt_buf;
 		if (session->outgoing) {
+			addr = &session->peer.ip;
 			assert ((task->pending_count == 1) && (task->pending[0] == session->handle));
 			task->pending_count = 0;
 			session_del_tasks(session, task);
 		}
 		/* Parse the packet and start resolving complete query */
-		int ret = parse_packet(pkt_buf);
+		int ret = parse_packet(pkt);
 		if (ret == 0) {
 			if (session->outgoing) {
 				/* To prevent slow lorris attack restart watchdog only after
@@ -2308,25 +2367,45 @@ int worker_process_tcp(struct worker_ctx *worker, uv_stream_t *handle,
 			} else {
 				/* Start only new queries,
 				 * not subrequests that are already pending */
-				ret = request_start(task->ctx, pkt_buf);
-				assert(ret == 0);
-				if (ret == 0) {
-					ret = qr_task_register(task, session);
+				ret = request_start(task->ctx, pkt);
+				if (ret != 0) {
+					/* Allocation of answer buffer has failed.
+					 * We can't notify client about failure,
+					 * so just end the task processing. */
+					qr_task_complete(task);
+					goto next_msg;
 				}
-				if (ret == 0) {
-					submitted += 1;
+
+				ret = qr_task_register(task, session);
+				if (ret != 0) {
+					/* Answer buffer has been allocated,
+					 * but task can't be attached to the given
+					 * session due to memory problems.
+					 * Finalize the task, otherwise it becomes orphaned. */
+					knot_pkt_init_response(task->ctx->req.answer, pkt);
+					qr_task_finalize(task, KR_STATE_FAIL);
+					goto next_msg;
 				}
+				submitted += 1;
 				if (task->leading) {
 					assert(false);
 				}
 			}
+		} else if (session->outgoing) {
+			/* Drop malformed packet and retry resolution */
+			pkt = NULL;
+			ret = 0;
+		} else {
+			qr_task_complete(task);
 		}
+		/* Only proceed if the message is valid, or it's an invalid response to
+		 * an outbound query which needs to be treated as a timeout. */
 		if (ret == 0) {
-			const struct sockaddr *addr = session->outgoing ? &session->peer.ip : NULL;
 			/* since there can be next dns message, we must to proceed
 			 * even if qr_task_step() returns error */
-			qr_task_step(task, addr, pkt_buf);
+			qr_task_step(task, addr, pkt);
 		}
+next_msg:
 		if (len > 0) {
 			/* TODO: this is simple via iteration; recursion doesn't really help */
 			ret = worker_process_tcp(worker, handle, msg, len);
@@ -2362,8 +2441,11 @@ struct qr_task *worker_resolve_start(struct worker_ctx *worker, knot_pkt_t *quer
 	/* Start task */
 	int ret = request_start(ctx, query);
 	if (ret != 0) {
+		/* task is attached to request context,
+		 * so dereference (and deallocate) it first */
+		request_del_tasks(ctx, task);
+		array_clear(ctx->tasks);
 		request_free(ctx);
-		qr_task_unref(task);
 		return NULL;
 	}
 
@@ -2481,6 +2563,11 @@ struct worker_ctx *worker_create(struct engine *engine, knot_mm_t *pool,
 	lua_setfield(engine->L, -2, "pid");
 	lua_pushnumber(engine->L, worker_count);
 	lua_setfield(engine->L, -2, "count");
+	/* Register table for worker per-request variables */
+	lua_newtable(engine->L);
+	lua_setfield(engine->L, -2, "vars");
+	lua_getfield(engine->L, -1, "vars");
+	worker->vars_table_ref = luaL_ref(engine->L, LUA_REGISTRYINDEX);
 	lua_pop(engine->L, 1);
 	return worker;
 }
