@@ -23,10 +23,10 @@
 #include "lib/layer/iterate.h"
 
 
-/** Reconstruct a name into a buffer (assuming length at least KNOT_DNAME_MAXLEN). */
+/** Reconstruct a name into a buffer (assuming length at least KNOT_DNAME_MAXLEN).
+ * \return kr_ok() or error code (<0). */
 static int dname_wire_reconstruct(knot_dname_t *buf, const struct key *k,
 		 knot_db_val_t kwz)
-/* TODO: probably move to a place shared with with NSEC3, perhaps with key_NSEC* */
 {
 	/* Reconstruct from key: first the ending, then zone name. */
 	int ret = knot_dname_lf2wire(buf, kwz.len, kwz.data);
@@ -127,15 +127,6 @@ static int kwz_between(knot_db_val_t k1, knot_db_val_t k2, knot_db_val_t k4)
 	}
 }
 
-static struct entry_h * entry_h_consistent_NSEC(knot_db_val_t data)
-{
-	/* ATM it's enough to just extend the checks for exact entries. */
-	const struct entry_h *eh = entry_h_consistent(data, KNOT_RRTYPE_NSEC);
-	bool ok = eh != NULL;
-	ok = ok && !(eh->is_packet || eh->has_ns || eh->has_cname || eh->has_dname
-		     || eh->has_optout);
-	return ok ? /*const-cast*/(struct entry_h *)eh : NULL;
-}
 
 /** NSEC1 range search.
  *
@@ -185,9 +176,10 @@ static const char * find_leq_NSEC1(struct kr_cache *cache, const struct kr_query
 		 * in case we searched before the very first one in the zone. */
 		return "range search found inconsistent entry";
 	}
-	/* FIXME(stale): passing just zone name instead of owner, as we don't
+	/* Passing just zone name instead of owner, as we don't
 	 * have it reconstructed at this point. */
-	int32_t new_ttl_ = get_new_ttl(eh, qry, k->zname, KNOT_RRTYPE_NSEC);
+	int32_t new_ttl_ = get_new_ttl(eh, qry, k->zname, KNOT_RRTYPE_NSEC,
+					qry->timestamp.tv_sec);
 	if (new_ttl_ < 0 || !kr_rank_test(eh->rank, KR_RANK_SECURE)) {
 		return "range search found stale or insecure entry";
 		/* TODO: remove the stale record *and* retry,
@@ -198,7 +190,7 @@ static const char * find_leq_NSEC1(struct kr_cache *cache, const struct kr_query
 	}
 	if (kwz_low) {
 		*kwz_low = (knot_db_val_t){
-			.data = key_nsec.data + nwz_off,
+			.data = (uint8_t *)key_nsec.data + nwz_off,
 			.len = key_nsec.len - nwz_off,
 		};	/* CACHE_KEY_DEF */
 	}
@@ -217,12 +209,12 @@ static const char * find_leq_NSEC1(struct kr_cache *cache, const struct kr_query
 	/* We know it starts before sname, so let's check the other end.
 	 * 1. construct the key for the next name - kwz_hi. */
 	/* it's *full* name ATM */
-	const knot_dname_t *next = eh->data + KR_CACHE_RR_COUNT_SIZE
-				 + 2 /* RDLENGTH from rfc1034 */;
-	if (!eh->data[0]) {
+	const knot_rdata_t *next = (const knot_rdata_t *)
+				(eh->data + KR_CACHE_RR_COUNT_SIZE);
+	if (KR_CACHE_RR_COUNT_SIZE != 2 || get_uint16(eh->data) == 0) {
 		assert(false);
 		return "ERROR";
-		/* TODO: more checks?  Also, `next` computation is kinda messy. */
+		/* TODO: more checks? */
 	}
 	/*
 	WITH_VERBOSE {
@@ -236,13 +228,18 @@ static const char * find_leq_NSEC1(struct kr_cache *cache, const struct kr_query
 		assert(false);
 		return "EINVAL";
 	}
-	ret = kr_dname_lf(chs, next, false);
-#if KNOT_VERSION_HEX >= ((2 << 16) | (7 << 8) | 0)
-	/* We have to lower-case it with libknot >= 2.7; see also RFC 6840 5.1. */
-	if (!ret) {
-		ret = knot_dname_to_lower(next);
+	{
+		/* Lower-case chs; see also RFC 6840 5.1.
+		 * LATER(optim.): we do lots of copying etc. */
+		knot_dname_t lower_buf[KNOT_DNAME_MAXLEN];
+		ret = knot_dname_to_wire(lower_buf, next->data,
+					 MIN(next->len, KNOT_DNAME_MAXLEN));
+		if (ret < 0) { /* _ESPACE */
+			return "range search found record with incorrect contents";
+		}
+		knot_dname_to_lower(lower_buf);
+		ret = kr_dname_lf(chs, lower_buf, false);
 	}
-#endif
 	if (ret) {
 		assert(false);
 		return "ERROR";
@@ -254,7 +251,7 @@ static const char * find_leq_NSEC1(struct kr_cache *cache, const struct kr_query
 	assert((ssize_t)(kwz_hi.len) >= 0);
 	/* 2. do the actual range check. */
 	const knot_db_val_t kwz_sname = {
-		.data = (void *)k->buf + 1 + nwz_off,
+		.data = (void *)/*const-cast*/(k->buf + 1 + nwz_off),
 		.len = k->buf[0] - k->zlf_len,
 	};
 	assert((ssize_t)(kwz_sname.len) >= 0);
@@ -297,9 +294,6 @@ int nsec1_encloser(struct key *k, struct answer *ans,
 		return ESKIP;
 	}
 
-	const struct entry_h *nsec_eh = val.data;
-	const void *nsec_eh_bound = val.data + val.len;
-
 	/* Get owner name of the record. */
 	const knot_dname_t *owner;
 	knot_dname_t owner_buf[KNOT_DNAME_MAXLEN];
@@ -312,20 +306,20 @@ int nsec1_encloser(struct key *k, struct answer *ans,
 	}
 	/* Basic checks OK -> materialize data. */
 	{
-		int ret = entry2answer(ans, AR_NSEC, nsec_eh, nsec_eh_bound,
+		const struct entry_h *nsec_eh = val.data;
+		int ret = entry2answer(ans, AR_NSEC, nsec_eh, knot_db_val_bound(val),
 					owner, KNOT_RRTYPE_NSEC, new_ttl);
 		if (ret) return kr_error(ret);
 	}
 
 	/* Final checks, split for matching vs. covering our sname. */
 	const knot_rrset_t *nsec_rr = ans->rrsets[AR_NSEC].set.rr;
-	uint8_t *bm = NULL;
-	uint16_t bm_size = 0;
-	knot_nsec_bitmap(&nsec_rr->rrs, &bm, &bm_size);
+	const uint8_t *bm = knot_nsec_bitmap(nsec_rr->rrs.rdata);
+	uint16_t bm_size = knot_nsec_bitmap_len(nsec_rr->rrs.rdata);
+	assert(bm);
 
 	if (exact_match) {
 		if (kr_nsec_bitmap_nodata_check(bm, bm_size, qry->stype, nsec_rr->owner) != 0) {
-			assert(bm);
 			VERBOSE_MSG(qry,
 				"=> NSEC sname: match but failed type check\n");
 			return ESKIP;
@@ -347,7 +341,7 @@ int nsec1_encloser(struct key *k, struct answer *ans,
 	/* NXDOMAIN proven *except* for wildcards. */
 	WITH_VERBOSE(qry) {
 		auto_free char *owner_str = kr_dname_text(nsec_rr->owner),
-			  *next_str = kr_dname_text(knot_nsec_next(&nsec_rr->rrs));
+			  *next_str = kr_dname_text(knot_nsec_next(nsec_rr->rrs.rdata));
 		VERBOSE_MSG(qry, "=> NSEC sname: covered by: %s -> %s, new TTL %d\n",
 				owner_str, next_str, new_ttl);
 	}
@@ -362,14 +356,12 @@ int nsec1_encloser(struct key *k, struct answer *ans,
 	 * LATER(optim.): it might be faster to use the LFs we already have.
 	 */
 	knot_dname_t next[KNOT_DNAME_MAXLEN];
-	int ret = knot_dname_to_wire(next, knot_nsec_next(&nsec_rr->rrs), sizeof(next));
-	if (ret >= 0) {
-		ret = knot_dname_to_lower(next);
-	}
+	int ret = knot_dname_to_wire(next, knot_nsec_next(nsec_rr->rrs.rdata), sizeof(next));
 	if (ret < 0) {
 		assert(!ret);
 		return kr_error(ret);
 	}
+	knot_dname_to_lower(next);
 	*clencl_labels = MAX(
 		nsec_matched,
 		knot_dname_matched_labels(qry->sname, next)
@@ -396,9 +388,8 @@ static bool nonexistence_ok(int cmp, const knot_rrset_t *rrs)
 	if (cmp != 2) {
 		return false;
 	}
-	uint8_t *bm = NULL;
-	uint16_t bm_size = 0;
-	knot_nsec_bitmap(&rrs->rrs, &bm, &bm_size);
+	const uint8_t *bm = knot_nsec_bitmap(rrs->rrs.rdata);
+	uint16_t bm_size = knot_nsec_bitmap_len(rrs->rrs.rdata);
 	return kr_nsec_children_in_zone_check(bm, bm_size) != 0;
 }
 
@@ -415,7 +406,7 @@ int nsec1_src_synth(struct key *k, struct answer *ans, const knot_dname_t *clenc
 	}
 	/* Check if our sname-covering NSEC also covers/matches SS. */
 	knot_db_val_t kwz = {
-		.data = key.data + nwz_off,
+		.data = (uint8_t *)key.data + nwz_off,
 		.len = key.len - nwz_off,
 	};
 	assert((ssize_t)(kwz.len) >= 0);
@@ -441,12 +432,11 @@ int nsec1_src_synth(struct key *k, struct answer *ans, const knot_dname_t *clenc
 			return kr_ok();
 		}
 		/* Materialize the record into answer (speculatively). */
-		const struct entry_h *nsec_eh = val.data;
-		const void *nsec_eh_bound = val.data + val.len;
 		knot_dname_t owner[KNOT_DNAME_MAXLEN];
 		int ret = dname_wire_reconstruct(owner, k, wild_low_kwz);
 		if (ret) return kr_error(ret);
-		ret = entry2answer(ans, AR_WILD, nsec_eh, nsec_eh_bound,
+		const struct entry_h *nsec_eh = val.data;
+		ret = entry2answer(ans, AR_WILD, nsec_eh, knot_db_val_bound(val),
 				   owner, KNOT_RRTYPE_NSEC, new_ttl);
 		if (ret) return kr_error(ret);
 		nsec_rr = ans->rrsets[AR_WILD].set.rr;
@@ -454,10 +444,9 @@ int nsec1_src_synth(struct key *k, struct answer *ans, const knot_dname_t *clenc
 
 	assert(nsec_rr);
 	const uint32_t new_ttl_log =
-		kr_verbose_status ? knot_rrset_ttl(nsec_rr) : -1;
-	uint8_t *bm = NULL;
-	uint16_t bm_size;
-	knot_nsec_bitmap(&nsec_rr->rrs, &bm, &bm_size);
+		kr_verbose_status ? nsec_rr->ttl : -1;
+	const uint8_t *bm = knot_nsec_bitmap(nsec_rr->rrs.rdata);
+	uint16_t bm_size = knot_nsec_bitmap_len(nsec_rr->rrs.rdata);
 	int ret;
 	struct answer_rrset * const arw = &ans->rrsets[AR_WILD];
 	if (!bm) {
@@ -481,7 +470,7 @@ int nsec1_src_synth(struct key *k, struct answer *ans, const knot_dname_t *clenc
 		/* We have a record proving wildcard non-existence. */
 		WITH_VERBOSE(qry) {
 			auto_free char *owner_str = kr_dname_text(nsec_rr->owner),
-				  *next_str = kr_dname_text(knot_nsec_next(&nsec_rr->rrs));
+				*next_str = kr_dname_text(knot_nsec_next(nsec_rr->rrs.rdata));
 			VERBOSE_MSG(qry, "=> NSEC wildcard: covered by: %s -> %s, new TTL %d\n",
 					owner_str, next_str, new_ttl_log);
 		}
@@ -513,7 +502,8 @@ int nsec1_src_synth(struct key *k, struct answer *ans, const knot_dname_t *clenc
 	ret = kr_ok();
 clean_wild:
 	if (arw->set.rr) { /* we may have matched AR_NSEC */
-		knot_rrset_free(&arw->set.rr, ans->mm);
+		knot_rrset_free(arw->set.rr, ans->mm);
+		arw->set.rr = NULL;
 		knot_rdataset_clear(&arw->sig_rds, ans->mm);
 	}
 	return ret;
