@@ -15,6 +15,7 @@
  */
 
 #include <assert.h>
+#include <stdlib.h>
 
 #include <libknot/descriptor.h>
 #include <libknot/dname.h>
@@ -22,7 +23,7 @@
 #include <libknot/rrset.h>
 #include <libknot/rrtype/nsec.h>
 #include <libknot/rrtype/rrsig.h>
-#include <dnssec/error.h>
+#include <libdnssec/error.h>
 
 #include "lib/defines.h"
 #include "lib/dnssec/nsec.h"
@@ -90,20 +91,15 @@ static int nsec_covers(const knot_rrset_t *nsec, const knot_dname_t *sname)
 	}
 
 	/* If NSEC 'owner' >= 'next', it means that there is nothing after 'owner' */
-#if KNOT_VERSION_HEX < ((2 << 16) | (7 << 8) | 0)
-	const knot_dname_t *next = knot_nsec_next(&nsec->rrs);
-#else
 	/* We have to lower-case it with libknot >= 2.7; see also RFC 6840 5.1. */
 	knot_dname_t next[KNOT_DNAME_MAXLEN];
-	int ret = knot_dname_to_wire(next, knot_nsec_next(&nsec->rrs), sizeof(next));
-	if (ret >= 0) {
-		ret = knot_dname_to_lower(next);
-	}
+	int ret = knot_dname_to_wire(next, knot_nsec_next(nsec->rrs.rdata), sizeof(next));
 	if (ret < 0) {
 		assert(!ret);
 		return kr_error(ret);
 	}
-#endif
+	knot_dname_to_lower(next);
+
 	const bool is_last_nsec = knot_dname_cmp(nsec->owner, next) >= 0;
 	const bool in_range = is_last_nsec || knot_dname_cmp(sname, next) < 0;
 	if (!in_range) {
@@ -113,12 +109,12 @@ static int nsec_covers(const knot_rrset_t *nsec, const knot_dname_t *sname)
 	 * sname might be under delegation from owner and thus
 	 * not in the zone of this NSEC at all.
 	 */
-	if (!knot_dname_is_sub(sname, nsec->owner)) {
+	if (knot_dname_in_bailiwick(sname, nsec->owner) <= 0) {
 		return kr_ok();
 	}
-	uint8_t *bm = NULL;
-	uint16_t bm_size = 0;
-	knot_nsec_bitmap(&nsec->rrs, &bm, &bm_size);
+	const uint8_t *bm = knot_nsec_bitmap(nsec->rrs.rdata);
+	uint16_t bm_size = knot_nsec_bitmap_len(nsec->rrs.rdata);
+
 	return kr_nsec_children_in_zone_check(bm, bm_size);
 }
 
@@ -239,15 +235,17 @@ static int coverign_rrsig_labels(const knot_rrset_t *nsec, const knot_pktsection
 			continue;
 		}
 
-		for (uint16_t j = 0; j < rrset->rrs.rr_count; ++j) {
-			if (knot_rrsig_type_covered(&rrset->rrs, j) != KNOT_RRTYPE_NSEC) {
+		knot_rdata_t *rdata_j = rrset->rrs.rdata;
+		for (uint16_t j = 0; j < rrset->rrs.count;
+				++j, rdata_j = knot_rdataset_next(rdata_j)) {
+			if (knot_rrsig_type_covered(rdata_j) != KNOT_RRTYPE_NSEC) {
 				continue;
 			}
 
 			if (ret < 0) {
-				ret = knot_rrsig_labels(&rrset->rrs, j);
+				ret = knot_rrsig_labels(rdata_j);
 			} else {
-				if (ret != knot_rrsig_labels(&rrset->rrs, j)) {
+				if (ret != knot_rrsig_labels(rdata_j)) {
 					return kr_error(EINVAL);
 				}
 			}
@@ -317,9 +315,8 @@ static int no_data_response_check_rrtype(int *flags, const knot_rrset_t *nsec,
 {
 	assert(flags && nsec);
 
-	uint8_t *bm = NULL;
-	uint16_t bm_size = 0;
-	knot_nsec_bitmap(&nsec->rrs, &bm, &bm_size);
+	const uint8_t *bm = knot_nsec_bitmap(nsec->rrs.rdata);
+	uint16_t bm_size = knot_nsec_bitmap_len(nsec->rrs.rdata);
 	int ret = kr_nsec_bitmap_nodata_check(bm, bm_size, type, nsec->owner);
 	if (ret == kr_ok()) {
 		*flags |= FLG_NOEXIST_RRTYPE;
@@ -380,13 +377,15 @@ static int wildcard_match_check(const knot_pkt_t *pkt, const knot_pktsection_t *
 		if (!knot_dname_is_wildcard(rrset->owner)) {
 			continue;
 		}
-		int wcard_labels = knot_dname_labels(rrset->owner, NULL);
-		int common_labels = knot_dname_matched_labels(rrset->owner, sname);
-		int rrsig_labels = coverign_rrsig_labels(rrset, sec);
-		if (wcard_labels < 1 ||
-		    common_labels != wcard_labels - 1 ||
-		    common_labels != rrsig_labels) {
-			continue;
+		if (!knot_dname_is_equal(rrset->owner, sname)) {
+			int wcard_labels = knot_dname_labels(rrset->owner, NULL);
+			int common_labels = knot_dname_matched_labels(rrset->owner, sname);
+			int rrsig_labels = coverign_rrsig_labels(rrset, sec);
+			if (wcard_labels < 1 ||
+			    common_labels != wcard_labels - 1 ||
+			    common_labels != rrsig_labels) {
+				continue;
+			}
 		}
 		int ret = no_data_response_check_rrtype(&flags, rrset, stype);
 		if (ret != 0) {
@@ -487,8 +486,6 @@ int kr_nsec_existence_denial(const knot_pkt_t *pkt, knot_section_t section_id,
 int kr_nsec_ref_to_unsigned(const knot_pkt_t *pkt)
 {
 	int nsec_found = 0;
-	uint8_t *bm = NULL;
-	uint16_t bm_size = 0;
 	const knot_pktsection_t *sec = knot_pkt_section(pkt, KNOT_AUTHORITY);
 	if (!sec) {
 		return kr_error(EINVAL);
@@ -518,7 +515,8 @@ int kr_nsec_ref_to_unsigned(const knot_pkt_t *pkt)
 				continue;
 			}
 			nsec_found = 1;
-			knot_nsec_bitmap(&nsec->rrs, &bm, &bm_size);
+			const uint8_t *bm = knot_nsec_bitmap(nsec->rrs.rdata);
+			uint16_t bm_size = knot_nsec_bitmap_len(nsec->rrs.rdata);
 			if (!bm) {
 				return kr_error(EINVAL);
 			}
@@ -560,9 +558,8 @@ int kr_nsec_matches_name_and_type(const knot_rrset_t *nsec,
 	if (!knot_dname_is_equal(nsec->owner, name)) {
 		return kr_error(ENOENT);
 	}
-	uint8_t *bm = NULL;
-	uint16_t bm_size = 0;
-	knot_nsec_bitmap(&nsec->rrs, &bm, &bm_size);
+	const uint8_t *bm = knot_nsec_bitmap(nsec->rrs.rdata);
+	uint16_t bm_size = knot_nsec_bitmap_len(nsec->rrs.rdata);
 	if (!bm) {
 		return kr_error(EINVAL);
 	}
