@@ -94,6 +94,7 @@ static int begin_yield(kr_layer_t *ctx) { return kr_ok(); }
 static int reset_yield(kr_layer_t *ctx) { return kr_ok(); }
 static int finish_yield(kr_layer_t *ctx) { return kr_ok(); }
 static int produce_yield(kr_layer_t *ctx, knot_pkt_t *pkt) { return kr_ok(); }
+static int checkout_yield(kr_layer_t *ctx, knot_pkt_t *packet, struct sockaddr *dst, int type) { return kr_ok(); }
 
 /** @internal Macro for iterating module layers. */
 #define RESUME_LAYERS(from, r, qry, func, ...) \
@@ -155,7 +156,7 @@ static int invalidate_ns(struct kr_rplan *rplan, struct kr_query *qry)
 		size_t addr_len = kr_inaddr_len(&qry->ns.addr[0].ip);
 		/* @warning _NOT_ thread-safe */
 		static knot_rdata_t rdata_arr[RDATA_ARR_MAX];
-		knot_rdata_init(rdata_arr, addr_len, (const uint8_t *)addr, 0);
+		knot_rdata_init(rdata_arr, addr_len, (const uint8_t *)addr);
 		return kr_zonecut_del(&qry->zone_cut, qry->ns.name, rdata_arr);
 	} else {
 		return kr_zonecut_del_all(&qry->zone_cut, qry->ns.name);
@@ -221,10 +222,12 @@ static int ns_fetch_cut(struct kr_query *qry, const knot_dname_t *requested_name
 		 * even if cut name is covered by TA. */
 		qry->flags.DNSSEC_WANT = false;
 		qry->flags.DNSSEC_INSECURE = true;
+		VERBOSE_MSG(qry, "=> going insecure because parent query is insecure\n");
 	} else if (kr_ta_covers_qry(req->ctx, qry->zone_cut.name, KNOT_RRTYPE_NS)) {
 		qry->flags.DNSSEC_WANT = true;
 	} else {
 		qry->flags.DNSSEC_WANT = false;
+		VERBOSE_MSG(qry, "=> going insecure because there's no covering TA\n");
 	}
 
 	struct kr_zonecut cut_found;
@@ -347,7 +350,7 @@ static int ns_resolve_addr(struct kr_query *qry, struct kr_request *param)
 			qry->flags.NO_MINIMIZE = true;
 			qry->flags.AWAIT_IPV6 = false;
 			qry->flags.AWAIT_IPV4 = false;
-			VERBOSE_MSG(qry, "=> circular dependepcy, retrying with non-minimized name\n");
+			VERBOSE_MSG(qry, "=> circular dependency, retrying with non-minimized name\n");
 		} else {
 			qry->ns.reputation |= KR_NS_NOIP4 | KR_NS_NOIP6;
 			kr_nsrep_update_rep(&qry->ns, qry->ns.reputation, ctx->cache_rep);
@@ -408,7 +411,7 @@ static int edns_erase_and_reserve(knot_pkt_t *pkt)
 	}
 
 	size_t len = knot_rrset_size(pkt->opt_rr);
-	int16_t rr_removed = pkt->opt_rr->rrs.rr_count;
+	int16_t rr_removed = pkt->opt_rr->rrs.count;
 	/* Decrease rrset counters. */
 	pkt->rrset_count -= 1;
 	pkt->sections[pkt->current].count -= 1;
@@ -463,10 +466,10 @@ static int answer_prepare(knot_pkt_t *answer, knot_pkt_t *query, struct kr_reque
 }
 
 /** @return error code, ignoring if forced to truncate the packet. */
-static int write_extra_records(const rr_array_t *arr, knot_pkt_t *answer)
+static int write_extra_records(const rr_array_t *arr, uint16_t reorder, knot_pkt_t *answer)
 {
 	for (size_t i = 0; i < arr->len; ++i) {
-		int err = knot_pkt_put(answer, 0, arr->at[i], 0);
+		int err = knot_pkt_put_rotate(answer, 0, arr->at[i], reorder, 0);
 		if (err != KNOT_EOK) {
 			return err == KNOT_ESPACE ? kr_ok() : kr_error(err);
 		}
@@ -480,8 +483,8 @@ static int write_extra_records(const rr_array_t *arr, knot_pkt_t *answer)
  * @param all_cname optionally output if all written RRs are CNAMEs and RRSIGs of CNAMEs
  * @return error code, ignoring if forced to truncate the packet.
  */
-static int write_extra_ranked_records(const ranked_rr_array_t *arr, knot_pkt_t *answer,
-				      bool *all_secure, bool *all_cname)
+static int write_extra_ranked_records(const ranked_rr_array_t *arr, uint16_t reorder,
+				      knot_pkt_t *answer, bool *all_secure, bool *all_cname)
 {
 	const bool has_dnssec = knot_pkt_has_dnssec(answer);
 	bool all_sec = true;
@@ -499,7 +502,7 @@ static int write_extra_ranked_records(const ranked_rr_array_t *arr, knot_pkt_t *
 				continue;
 			}
 		}
-		err = knot_pkt_put(answer, 0, rr, 0);
+		err = knot_pkt_put_rotate(answer, 0, rr, reorder, 0);
 		if (err != KNOT_EOK) {
 			if (err == KNOT_ESPACE) {
 				err = kr_ok();
@@ -535,12 +538,7 @@ static int answer_padding(struct kr_request *request)
 	int32_t pad_bytes = -1;
 
 	if (padding == -1) { /* use the default padding policy from libknot */
-#if KNOT_VERSION_HEX < ((2 << 16) | (4 << 8) | 3)
-		/* no knot_edns_default_padding_size available in libknot */
-		padding = KR_DEFAULT_TLS_PADDING;
-#else
-		pad_bytes =  knot_edns_default_padding_size(answer, opt_rr);
-#endif
+		pad_bytes =  knot_pkt_default_padding_size(answer, opt_rr);
 	}
 	if (padding >= 2) {
 		int32_t max_pad_bytes = knot_edns_get_payload(opt_rr) - (answer->size + knot_rrset_size(opt_rr));
@@ -596,8 +594,8 @@ static int answer_finalize(struct kr_request *request, int state)
 	/* AD flag.  We can only change `secure` from true to false.
 	 * Be conservative.  Primary approach: check ranks of all RRs in wire.
 	 * Only "negative answers" need special handling. */
-	bool secure = (last != NULL); /* suspicious otherwise */
-	VERBOSE_MSG(NULL, "AD: secure (start)\n");
+	bool secure = last != NULL && state == KR_STATE_DONE /*< suspicious otherwise */
+		&& knot_pkt_qtype(answer) != KNOT_RRTYPE_RRSIG;
 	if (last && (last->flags.STUB)) {
 		secure = false; /* don't trust forwarding for now */
 	}
@@ -606,6 +604,7 @@ static int answer_finalize(struct kr_request *request, int state)
 		secure = false; /* the last answer is insecure due to opt-out */
 	}
 
+	const uint16_t reorder = last ? last->reorder : 0;
 	bool answ_all_cnames = false/*arbitrary*/;
 	if (request->answ_selected.len > 0) {
 		assert(answer->current <= KNOT_ANSWER);
@@ -613,24 +612,24 @@ static int answer_finalize(struct kr_request *request, int state)
 		if (answer->current < KNOT_ANSWER) {
 			knot_pkt_begin(answer, KNOT_ANSWER);
 		}
-		if (write_extra_ranked_records(&request->answ_selected, answer,
-						&secure, &answ_all_cnames))
+		if (write_extra_ranked_records(&request->answ_selected, reorder,
+						answer, &secure, &answ_all_cnames))
 		{
 			return answer_fail(request);
 		}
 	}
 
-	VERBOSE_MSG(NULL, "AD: secure (between ANS and AUTH)\n");
 	/* Write authority records. */
 	if (answer->current < KNOT_AUTHORITY) {
 		knot_pkt_begin(answer, KNOT_AUTHORITY);
 	}
-	if (write_extra_ranked_records(&request->auth_selected, answer, &secure, NULL)) {
+	if (write_extra_ranked_records(&request->auth_selected, reorder,
+	    answer, &secure, NULL)) {
 		return answer_fail(request);
 	}
 	/* Write additional records. */
 	knot_pkt_begin(answer, KNOT_ADDITIONAL);
-	if (write_extra_records(&request->additional, answer)) {
+	if (write_extra_records(&request->additional, reorder, answer)) {
 		return answer_fail(request);
 	}
 	/* Write EDNS information */
@@ -645,35 +644,38 @@ static int answer_finalize(struct kr_request *request, int state)
 		ret = edns_put(answer);
 	}
 
+	if (!last) secure = false; /*< should be no-op, mostly documentation */
 	/* AD: "negative answers" need more handling. */
-	if (last && secure) {
-		VERBOSE_MSG(NULL, "AD: secure (1)\n");
-		if (kr_response_classify(answer) != PKT_NOERROR
-		    /* Additionally check for CNAME chains that "end in NODATA",
-		     * as those would also be PKT_NOERROR. */
-		    || (answ_all_cnames && knot_pkt_qtype(answer) != KNOT_RRTYPE_CNAME))
-		{
-			secure = secure && last->flags.DNSSEC_WANT
-				&& !last->flags.DNSSEC_BOGUS && !last->flags.DNSSEC_INSECURE;
-		}
-	}
-	/* Clear AD if not secure.  ATM answer has AD=1 if requested secured answer. */
-	if (!secure || state != KR_STATE_DONE
-	    || knot_pkt_qtype(answer) == KNOT_RRTYPE_RRSIG) {
-		knot_wire_clear_ad(answer->wire);
+	if (kr_response_classify(answer) != PKT_NOERROR
+	    /* Additionally check for CNAME chains that "end in NODATA",
+	     * as those would also be PKT_NOERROR. */
+	    || (answ_all_cnames && knot_pkt_qtype(answer) != KNOT_RRTYPE_CNAME)) {
+
+		secure = secure && last->flags.DNSSEC_WANT
+			&& !last->flags.DNSSEC_BOGUS && !last->flags.DNSSEC_INSECURE;
 	}
 
-	if (last) {
+	if (secure) {
 		struct kr_query *cname_parent = last->cname_parent;
 		while (cname_parent != NULL) {
 			if (cname_parent->flags.DNSSEC_OPTOUT) {
-				knot_wire_clear_ad(answer->wire);
+				secure = false;
 				break;
 			}
 			cname_parent = cname_parent->cname_parent;
 		}
 	}
 
+	/* No detailed analysis ATM, just _SECURE or not.
+	 * LATER: request->rank might better be computed in validator's finish phase. */
+	VERBOSE_MSG(NULL, "  AD: request%s classified as SECURE\n", secure ? "" : " NOT");
+	request->rank = secure ? KR_RANK_SECURE : KR_RANK_INITIAL;
+
+	/* Clear AD if not secure.  ATM answer has AD=1 if requested secured answer. */
+	if (!secure) {
+		knot_wire_clear_ad(answer->wire);
+	}
+
 	return ret;
 }
 
@@ -726,6 +728,7 @@ int kr_resolve_begin(struct kr_request *request, struct kr_context *ctx, knot_pk
 	array_init(request->add_selected);
 	request->answ_validated = false;
 	request->auth_validated = false;
+	request->rank = KR_RANK_INITIAL;
 	request->trace_log = NULL;
 	request->trace_finish = NULL;
 
@@ -750,7 +753,7 @@ static int resolve_query(struct kr_request *request, const knot_pkt_t *packet)
 	} else if (cookie_ctx && cookie_ctx->srvr.enabled &&
 		   knot_wire_get_qdcount(packet->wire) == 0 &&
 		   knot_pkt_has_edns(packet) &&
-		   knot_edns_has_option(packet->opt_rr, KNOT_EDNS_OPTION_COOKIE)) {
+		   knot_pkt_edns_option(packet, KNOT_EDNS_OPTION_COOKIE)) {
 		/* Plan empty query only for cookies. */
 		qry = kr_rplan_push_empty(rplan, NULL);
 	}
@@ -921,7 +924,7 @@ int kr_resolve_consume(struct kr_request *request, const struct sockaddr *src, k
 		}
 	} else {
 		/* Packet cleared, derandomize QNAME. */
-		knot_dname_t *qname_raw = (knot_dname_t *)knot_pkt_qname(packet);
+		knot_dname_t *qname_raw = knot_pkt_qname(packet);
 		if (qname_raw && qry->secret != 0) {
 			randomized_qname_case(qname_raw, qry->secret);
 		}
@@ -1002,7 +1005,7 @@ static int forward_trust_chain_check(struct kr_request *request, struct kr_query
 	if (qry->parent != NULL &&
 	    !(qry->forward_flags.CNAME) &&
 	    !(qry->flags.DNS64_MARK) &&
-	    knot_dname_in(qry->parent->zone_cut.name, qry->zone_cut.name)) {
+	    knot_dname_in_bailiwick(qry->zone_cut.name, qry->parent->zone_cut.name) >= 0) {
 		return KR_STATE_PRODUCE;
 	}
 
@@ -1144,8 +1147,7 @@ static int forward_trust_chain_check(struct kr_request *request, struct kr_query
 		qry->flags.DNSSEC_WANT = true;
 		want_secured = true;
 		WITH_VERBOSE(qry) {
-		char qname_str[KNOT_DNAME_MAXLEN];
-		knot_dname_to_str(qname_str, wanted_name, sizeof(qname_str));
+		KR_DNAME_GET_STR(qname_str, wanted_name);
 		VERBOSE_MSG(qry, ">< TA: '%s'\n", qname_str);
 		}
 	}
@@ -1223,8 +1225,7 @@ static int trust_chain_check(struct kr_request *request, struct kr_query *qry)
 			qry->zone_cut.trust_anchor = knot_rrset_copy(ta_rr, qry->zone_cut.pool);
 
 			WITH_VERBOSE(qry) {
-			char qname_str[KNOT_DNAME_MAXLEN];
-			knot_dname_to_str(qname_str, ta_rr->owner, sizeof(qname_str));
+			KR_DNAME_GET_STR(qname_str, ta_rr->owner);
 			VERBOSE_MSG(qry, ">< TA: '%s'\n", qname_str);
 			}
 		}
@@ -1299,7 +1300,8 @@ static int zone_cut_check(struct kr_request *request, struct kr_query *qry, knot
 	 * (and need glue from parent), or DS refetch. */
 	if (qry->parent) {
 		const knot_dname_t *parent = qry->parent->zone_cut.name;
-		if (parent[0] != '\0' && knot_dname_in(parent, qry->sname)) {
+		if (parent[0] != '\0'
+		    && knot_dname_in_bailiwick(qry->sname, parent) >= 0) {
 			requested_name = knot_wire_next_label(parent, NULL);
 		}
 	} else if ((qry->stype == KNOT_RRTYPE_DS) && (qry->sname[0] != '\0')) {
@@ -1369,10 +1371,12 @@ int kr_resolve_produce(struct kr_request *request, struct sockaddr **dst, int *t
 		/* Caller is interested in always tracking a zone cut, even if the answer is cached
 		 * this is normally not required, and incurrs another cache lookups for cached answer. */
 		if (qry->flags.ALWAYS_CUT) {
-			switch(zone_cut_check(request, qry, packet)) {
-			case KR_STATE_FAIL: return KR_STATE_FAIL;
-			case KR_STATE_DONE: return KR_STATE_PRODUCE;
-			default: break;
+			if (!(qry->flags.STUB)) {
+				switch(zone_cut_check(request, qry, packet)) {
+				case KR_STATE_FAIL: return KR_STATE_FAIL;
+				case KR_STATE_DONE: return KR_STATE_PRODUCE;
+				default: break;
+				}
 			}
 		}
 		/* Resolve current query and produce dependent or finish */
@@ -1470,7 +1474,7 @@ ns_election:
 	/* Randomize query case (if not in safemode or turned off) */
 	qry->secret = (qry->flags.SAFEMODE || qry->flags.NO_0X20)
 			? 0 : kr_rand_uint(0);
-	knot_dname_t *qname_raw = (knot_dname_t *)knot_pkt_qname(packet);
+	knot_dname_t *qname_raw = knot_pkt_qname(packet);
 	randomized_qname_case(qname_raw, qry->secret);
 
 	/*
@@ -1532,6 +1536,14 @@ int kr_resolve_checkout(struct kr_request *request, struct sockaddr *src,
 	}
 	struct kr_query *qry = array_tail(rplan->pending);
 
+	/* Run the checkout layers and cancel on failure. */
+	int state = request->state;
+	ITERATE_LAYERS(request, qry, checkout, packet, dst, type);
+	if (request->state == KR_STATE_FAIL) {
+		request->state = state; /* Restore */
+		return kr_error(ECANCELED);
+	}
+
 #if defined(ENABLE_COOKIES)
 	/* Update DNS cookies in request. */
 	if (type == SOCK_DGRAM) { /* @todo: Add cookies also over TCP? */
@@ -1552,10 +1564,13 @@ int kr_resolve_checkout(struct kr_request *request, struct sockaddr *src,
 	}
 
 	WITH_VERBOSE(qry) {
-	char qname_str[KNOT_DNAME_MAXLEN], zonecut_str[KNOT_DNAME_MAXLEN], ns_str[INET6_ADDRSTRLEN], type_str[16];
-	knot_dname_to_str(qname_str, knot_pkt_qname(packet), sizeof(qname_str));
-	knot_dname_to_str(zonecut_str, qry->zone_cut.name, sizeof(zonecut_str));
-	knot_rrtype_to_string(knot_pkt_qtype(packet), type_str, sizeof(type_str));
+
+	char ns_str[INET6_ADDRSTRLEN];
+
+	KR_DNAME_GET_STR(qname_str, knot_pkt_qname(packet));
+	KR_DNAME_GET_STR(zonecut_str, qry->zone_cut.name);
+	KR_RRTYPE_GET_STR(type_str, knot_pkt_qtype(packet));
+
 	for (size_t i = 0; i < KR_NSREP_MAXADDR; ++i) {
 		struct sockaddr *addr = &qry->ns.addr[i].ip;
 		if (addr->sa_family == AF_UNSPEC) {
