@@ -179,9 +179,7 @@ static struct kr_query *kr_rplan_push_query(struct kr_rplan *rplan,
 	qry->timestamp_mono = kr_now();
 	qry->creation_time_mono = parent ? parent->creation_time_mono : qry->timestamp_mono;
 	kr_zonecut_init(&qry->zone_cut, (const uint8_t *)"", rplan->pool);
-	qry->reorder = qry->flags.REORDER_RR
-		? knot_wire_get_id(rplan->request->answer->wire)
-		: 0;
+	qry->reorder = qry->flags.REORDER_RR ? kr_rand_uint(0) : 0;
 
 	/* When forwarding, keep the nameserver addresses. */
 	if (parent && parent->flags.FORWARD && qry->flags.FORWARD) {
@@ -230,9 +228,8 @@ struct kr_query *kr_rplan_push(struct kr_rplan *rplan, struct kr_query *parent,
 	qry->stype = type;
 
 	WITH_VERBOSE(qry) {
-	char name_str[KNOT_DNAME_MAXLEN], type_str[16];
-	knot_dname_to_str(name_str, name, sizeof(name_str));
-	knot_rrtype_to_string(type, type_str, sizeof(type_str));
+	KR_DNAME_GET_STR(name_str, name);
+	KR_RRTYPE_GET_STR(type_str, type);
 	VERBOSE_MSG(parent, "plan '%s' type '%s'\n", name_str, type_str);
 	}
 	return qry;
