@@ -261,34 +261,43 @@ int kr_memreserve(void *baton, char **mem, size_t elm_size, size_t want, size_t
     return -1;
 }
 
-int kr_pkt_recycle(knot_pkt_t *pkt)
+static int pkt_recycle(knot_pkt_t *pkt, bool keep_question)
 {
-	pkt->rrset_count = 0;
-	pkt->size = KNOT_WIRE_HEADER_SIZE;
-	pkt->current = KNOT_ANSWER;
-	knot_wire_set_qdcount(pkt->wire, 0);
+	/* The maximum size of a header + query name + (class, type) */
+	uint8_t buf[KNOT_WIRE_HEADER_SIZE + KNOT_DNAME_MAXLEN + 2 * sizeof(uint16_t)];
+
+	/* Save header and the question section */
+	size_t base_size = KNOT_WIRE_HEADER_SIZE;
+	if (keep_question) {
+		base_size += knot_pkt_question_size(pkt);
+	}
+	assert(base_size <= sizeof(buf));
+	memcpy(buf, pkt->wire, base_size);
+
+	/* Clear the packet and its auxiliary structures */
+	knot_pkt_clear(pkt);
+
+	/* Restore header and question section and clear counters */
+	pkt->size = base_size;
+	memcpy(pkt->wire, buf, base_size);
+	knot_wire_set_qdcount(pkt->wire, keep_question);
 	knot_wire_set_ancount(pkt->wire, 0);
 	knot_wire_set_nscount(pkt->wire, 0);
 	knot_wire_set_arcount(pkt->wire, 0);
-	memset(pkt->sections, 0, sizeof(pkt->sections));
+
+	/* Reparse question */
 	knot_pkt_begin(pkt, KNOT_ANSWER);
 	return knot_pkt_parse_question(pkt);
 }
 
+int kr_pkt_recycle(knot_pkt_t *pkt)
+{
+	return pkt_recycle(pkt, false);
+}
+
 int kr_pkt_clear_payload(knot_pkt_t *pkt)
 {
-	pkt->rrset_count = 0;
-	pkt->size = KNOT_WIRE_HEADER_SIZE + pkt->qname_size +
-		    2 * sizeof(uint16_t); /* QTYPE + QCLASS */
-	pkt->parsed = KNOT_WIRE_HEADER_SIZE;
-	pkt->current = KNOT_ANSWER;
-	knot_wire_set_ancount(pkt->wire, 0);
-	knot_wire_set_nscount(pkt->wire, 0);
-	knot_wire_set_arcount(pkt->wire, 0);
-	memset(&pkt->sections[KNOT_ANSWER], 0, sizeof(knot_pktsection_t) *
-	       (KNOT_PKT_SECTIONS - (KNOT_ANSWER + 1)));
-	knot_pkt_begin(pkt, KNOT_ANSWER);
-	return knot_pkt_parse_question(pkt);
+	return pkt_recycle(pkt, knot_wire_get_qdcount(pkt->wire));
 }
 
 int kr_pkt_put(knot_pkt_t *pkt, const knot_dname_t *name, uint32_t ttl,
@@ -299,12 +308,12 @@ int kr_pkt_put(knot_pkt_t *pkt, const knot_dname_t *name, uint32_t ttl,
 	}
 	/* Create empty RR */
 	knot_rrset_t rr;
-	knot_rrset_init(&rr, knot_dname_copy(name, &pkt->mm), rtype, rclass);
+	knot_rrset_init(&rr, knot_dname_copy(name, &pkt->mm), rtype, rclass, ttl);
 	/* Create RDATA
 	 * @warning _NOT_ thread safe.
 	 */
 	static knot_rdata_t rdata_arr[RDATA_ARR_MAX];
-	knot_rdata_init(rdata_arr, rdlen, rdata, ttl);
+	knot_rdata_init(rdata_arr, rdlen, rdata);
 	knot_rdataset_add(&rr.rrs, rdata_arr, &pkt->mm);
 	/* Append RR */
 	return knot_pkt_put(pkt, 0, &rr, KNOT_PF_FREE);
@@ -356,6 +365,39 @@ int kr_sockaddr_len(const struct sockaddr *addr)
 	}
 }
 
+int kr_sockaddr_cmp(const struct sockaddr *left, const struct sockaddr *right)
+{
+	if (!left || !right) {
+		return kr_error(EINVAL);
+	}
+	if (left->sa_family != right->sa_family) {
+		return kr_error(EFAULT);
+	}
+	if (left->sa_family == AF_INET) {
+		struct sockaddr_in *left_in = (struct sockaddr_in *)left;
+		struct sockaddr_in *right_in = (struct sockaddr_in *)right;
+		if (left_in->sin_addr.s_addr != right_in->sin_addr.s_addr) {
+			return kr_error(EFAULT);
+		}
+		if (left_in->sin_port != right_in->sin_port) {
+			return kr_error(EFAULT);
+		}
+	} else if (left->sa_family == AF_INET6) {
+		struct sockaddr_in6 *left_in6 = (struct sockaddr_in6 *)left;
+		struct sockaddr_in6 *right_in6 = (struct sockaddr_in6 *)right;
+		if (memcmp(&left_in6->sin6_addr, &right_in6->sin6_addr,
+			   sizeof(struct in6_addr)) != 0) {
+			return kr_error(EFAULT);
+		}
+		if (left_in6->sin6_port != right_in6->sin6_port) {
+			return kr_error(EFAULT);
+		}
+	} else {
+		return kr_error(ENOENT);
+	}
+	return kr_ok();
+}
+
 uint16_t kr_inaddr_port(const struct sockaddr *addr)
 {
 	if (!addr) {
@@ -617,8 +659,8 @@ static inline bool rrsets_match(const knot_rrset_t *rr1, const knot_rrset_t *rr2
 {
 	bool match = rr1->type == rr2->type && rr1->rclass == rr2->rclass;
 	if (match && rr2->type == KNOT_RRTYPE_RRSIG) {
-		match = match && knot_rrsig_type_covered(&rr1->rrs, 0)
-				  == knot_rrsig_type_covered(&rr2->rrs, 0);
+		match = match && knot_rrsig_type_covered(rr1->rrs.rdata)
+				  == knot_rrsig_type_covered(rr2->rrs.rdata);
 	}
 	match = match && knot_dname_is_equal(rr1->owner, rr2->owner);
 	return match;
@@ -840,9 +882,7 @@ char *kr_pkt_text(const knot_pkt_t *pkt)
 	static const char * snames[] = {
 		";; ANSWER SECTION", ";; AUTHORITY SECTION", ";; ADDITIONAL SECTION"
 	};
-	char rrtype[32];
 	char flags[32];
-	char qname[KNOT_DNAME_MAXLEN];
 	uint8_t pkt_rcode = knot_wire_get_rcode(pkt->wire);
 	uint8_t pkt_opcode = knot_wire_get_opcode(pkt->wire);
 	const char *rcode_str = "Unknown";
@@ -876,8 +916,8 @@ char *kr_pkt_text(const knot_pkt_t *pkt)
 	}
 
 	if (qdcount == 1) {
-		knot_dname_to_str(qname, knot_pkt_qname(pkt), KNOT_DNAME_MAXLEN);
-		knot_rrtype_to_string(knot_pkt_qtype(pkt), rrtype, sizeof(rrtype));
+		KR_DNAME_GET_STR(qname, knot_pkt_qname(pkt));
+		KR_RRTYPE_GET_STR(rrtype, knot_pkt_qtype(pkt));
 		ptr = mp_printf_append(mp, ptr, ";; QUESTION SECTION\n%s\t\t%s\n", qname, rrtype);
 	} else if (qdcount > 1) {
 		ptr = mp_printf_append(mp, ptr, ";; Warning: unsupported QDCOUNT %hu\n", qdcount);
@@ -970,3 +1010,30 @@ finish:
 	++d;
 	return d - dst;
 }
+
+void kr_rrset_init(knot_rrset_t *rrset, knot_dname_t *owner,
+			uint16_t type, uint16_t rclass, uint32_t ttl)
+{
+	assert(rrset);
+	knot_rrset_init(rrset, owner, type, rclass, ttl);
+}
+uint16_t kr_pkt_qclass(const knot_pkt_t *pkt)
+{
+	return knot_pkt_qclass(pkt);
+}
+uint16_t kr_pkt_qtype(const knot_pkt_t *pkt)
+{
+	return knot_pkt_qtype(pkt);
+}
+uint32_t kr_rrsig_sig_inception(const knot_rdata_t *rdata)
+{
+	return knot_rrsig_sig_inception(rdata);
+}
+uint32_t kr_rrsig_sig_expiration(const knot_rdata_t *rdata)
+{
+	return knot_rrsig_sig_expiration(rdata);
+}
+uint16_t kr_rrsig_type_covered(const knot_rdata_t *rdata)
+{
+	return knot_rrsig_type_covered(rdata);
+}
