@@ -23,7 +23,6 @@
 
 #include <libknot/packet/pkt.h>
 #include <libknot/descriptor.h>
-#include <libknot/rrtype/aaaa.h>
 #include <ccan/json/json.h>
 #include <ucw/mempool.h>
 #include <contrib/cleanup.h>
@@ -41,6 +40,7 @@
 struct hints_data {
 	struct kr_zonecut hints;
 	struct kr_zonecut reverse_hints;
+	bool use_nodata;
 };
 
 /** Useful for returning from module properties. */
@@ -52,17 +52,23 @@ static char * bool2jsonstr(bool val)
 	return result;
 }
 
-static int put_answer(knot_pkt_t *pkt, knot_rrset_t *rr)
+static int put_answer(knot_pkt_t *pkt, struct kr_query *qry, knot_rrset_t *rr, bool use_nodata)
 {
 	int ret = 0;
-	if (!knot_rrset_empty(rr)) {
+	if (!knot_rrset_empty(rr) || use_nodata) {
 		/* Update packet question */
 		if (!knot_dname_is_equal(knot_pkt_qname(pkt), rr->owner)) {
 			kr_pkt_recycle(pkt);
-			knot_pkt_put_question(pkt, rr->owner, rr->rclass, rr->type);
+			knot_pkt_put_question(pkt, qry->sname, qry->sclass, qry->stype);
+		}
+		if (!knot_rrset_empty(rr)) {
+			/* Append to packet */
+			ret = knot_pkt_put_rotate(pkt, KNOT_COMPR_HINT_QNAME, rr,
+						  qry->reorder, KNOT_PF_FREE);
+		} else {
+			/* Return empty answer if name exists, but type doesn't match */
+			knot_wire_set_aa(pkt->wire);
 		}
-		/* Append to packet */
-		ret = knot_pkt_put(pkt, KNOT_COMPR_HINT_QNAME, rr, KNOT_PF_FREE);
 	} else {
 		ret = kr_error(ENOENT);
 	}
@@ -73,7 +79,7 @@ static int put_answer(knot_pkt_t *pkt, knot_rrset_t *rr)
 	return ret;
 }
 
-static int satisfy_reverse(struct kr_zonecut *hints, knot_pkt_t *pkt, struct kr_query *qry)
+static int satisfy_reverse(struct kr_zonecut *hints, knot_pkt_t *pkt, struct kr_query *qry, bool use_nodata)
 {
 	/* Find a matching name */
 	pack_t *addr_set = kr_zonecut_find(hints, qry->sname);
@@ -82,20 +88,20 @@ static int satisfy_reverse(struct kr_zonecut *hints, knot_pkt_t *pkt, struct kr_
 	}
 	knot_dname_t *qname = knot_dname_copy(qry->sname, &pkt->mm);
 	knot_rrset_t rr;
-	knot_rrset_init(&rr, qname, KNOT_RRTYPE_PTR, KNOT_CLASS_IN);
+	knot_rrset_init(&rr, qname, KNOT_RRTYPE_PTR, KNOT_CLASS_IN, 0);
 
 	/* Append address records from hints */
 	uint8_t *addr = pack_last(*addr_set);
 	if (addr != NULL) {
 		size_t len = pack_obj_len(addr);
 		void *addr_val = pack_obj_val(addr);
-		knot_rrset_add_rdata(&rr, addr_val, len, 0, &pkt->mm);
+		knot_rrset_add_rdata(&rr, addr_val, len, &pkt->mm);
 	}
 
-	return put_answer(pkt, &rr);
+	return put_answer(pkt, qry, &rr, use_nodata);
 }
 
-static int satisfy_forward(struct kr_zonecut *hints, knot_pkt_t *pkt, struct kr_query *qry)
+static int satisfy_forward(struct kr_zonecut *hints, knot_pkt_t *pkt, struct kr_query *qry, bool use_nodata)
 {
 	/* Find a matching name */
 	pack_t *addr_set = kr_zonecut_find(hints, qry->sname);
@@ -104,7 +110,7 @@ static int satisfy_forward(struct kr_zonecut *hints, knot_pkt_t *pkt, struct kr_
 	}
 	knot_dname_t *qname = knot_dname_copy(qry->sname, &pkt->mm);
 	knot_rrset_t rr;
-	knot_rrset_init(&rr, qname, qry->stype, qry->sclass);
+	knot_rrset_init(&rr, qname, qry->stype, qry->sclass, 0);
 	size_t family_len = sizeof(struct in_addr);
 	if (rr.type == KNOT_RRTYPE_AAAA) {
 		family_len = sizeof(struct in6_addr);
@@ -116,12 +122,12 @@ static int satisfy_forward(struct kr_zonecut *hints, knot_pkt_t *pkt, struct kr_
 		size_t len = pack_obj_len(addr);
 		void *addr_val = pack_obj_val(addr);
 		if (len == family_len) {
-			knot_rrset_add_rdata(&rr, addr_val, len, 0, &pkt->mm);
+			knot_rrset_add_rdata(&rr, addr_val, len, &pkt->mm);
 		}
 		addr = pack_obj_next(addr);
 	}
 
-	return put_answer(pkt, &rr);
+	return put_answer(pkt, qry, &rr, use_nodata);
 }
 
 static int query(kr_layer_t *ctx, knot_pkt_t *pkt)
@@ -141,11 +147,11 @@ static int query(kr_layer_t *ctx, knot_pkt_t *pkt)
 	switch(qry->stype) {
 	case KNOT_RRTYPE_A:
 	case KNOT_RRTYPE_AAAA: /* Find forward record hints */
-		if (satisfy_forward(&data->hints, pkt, qry) != 0)
+		if (satisfy_forward(&data->hints, pkt, qry, data->use_nodata) != 0)
 			return ctx->state;
 		break;
 	case KNOT_RRTYPE_PTR: /* Find PTR record */
-		if (satisfy_reverse(&data->reverse_hints, pkt, qry) != 0)
+		if (satisfy_reverse(&data->reverse_hints, pkt, qry, data->use_nodata) != 0)
 			return ctx->state;
 		break;
 	default:
@@ -185,7 +191,7 @@ static const knot_rdata_t * addr2rdata(const char *addr) {
 	static knot_rdata_t rdata_arr[RDATA_ARR_MAX];
 	size_t addr_len = kr_inaddr_len((struct sockaddr *)&ss);
 	const uint8_t *raw_addr = (const uint8_t *)kr_inaddr((struct sockaddr *)&ss);
-	knot_rdata_init(rdata_arr, addr_len, raw_addr, 0);
+	knot_rdata_init(rdata_arr, addr_len, raw_addr);
 	return rdata_arr;
 }
 
@@ -248,10 +254,7 @@ static int add_pair(struct kr_zonecut *hints, const char *name, const char *addr
 	if (!knot_dname_from_str(key, name, sizeof(key))) {
 		return kr_error(EINVAL);
 	}
-	int ret = knot_dname_to_lower(key);
-	if (ret) {
-		return ret;
-	}
+	knot_dname_to_lower(key);
 	const knot_rdata_t *rdata = addr2rdata(addr);
 	if (!rdata) {
 		return kr_error(EINVAL);
@@ -275,7 +278,7 @@ static int add_reverse_pair(struct kr_zonecut *hints, const char *name, const ch
 
 	/* Build RDATA */
 	knot_rdata_t rdata[RDATA_ARR_MAX];
-	knot_rdata_init(rdata, knot_dname_size(ptr_name), ptr_name, 0);
+	knot_rdata_init(rdata, knot_dname_size(ptr_name), ptr_name);
 
 	return kr_zonecut_add(hints, key, rdata);
 }
@@ -292,7 +295,7 @@ static int del_pair(struct hints_data *data, const char *name, const char *addr)
 		return kr_error(EINVAL);
 	}
 	knot_rdata_t ptr_rdata[RDATA_ARR_MAX];
-	knot_rdata_init(ptr_rdata, knot_dname_size(key), key, 0);
+	knot_rdata_init(ptr_rdata, knot_dname_size(key), key);
 
         if (addr) {
 		/* Remove the pair. */
@@ -512,11 +515,7 @@ static char* pack_hints(struct kr_zonecut *hints) {
 	JsonNode *root_node = json_mkobject();
 	trie_it_t *it;
 	for (it = trie_it_begin(hints->nsset); !trie_it_finished(it); trie_it_next(it)) {
-		char nsname_str[KNOT_DNAME_MAXLEN] = {'\0'};
-		knot_dname_to_str(nsname_str,
-					/* here we trust that it's a correct dname */
-					(const knot_dname_t *)trie_it_key(it, NULL),
-					sizeof(nsname_str));
+		KR_DNAME_GET_STR(nsname_str, (const knot_dname_t *)trie_it_key(it, NULL));
 		JsonNode *addr_list = pack_addrs((pack_t *)*trie_it_val(it));
 		if (!addr_list) goto error;
 		json_append_member(root_node, nsname_str, addr_list);
@@ -574,6 +573,22 @@ static char* hint_root_file(void *env, struct kr_module *module, const char *arg
 	return strdup(err_msg ? err_msg : "");
 }
 
+static char* hint_use_nodata(void *env, struct kr_module *module, const char *args)
+{
+	struct hints_data *data = module->data;
+	if (!args) {
+		return NULL;
+	}
+
+	JsonNode *root_node = json_decode(args);
+	if (!root_node || root_node->tag != JSON_BOOL) {
+		return bool2jsonstr(false);
+	}
+
+	data->use_nodata = root_node->bool_;
+	return bool2jsonstr(true);
+}
+
 /*
  * Module implementation.
  */
@@ -660,6 +675,7 @@ struct kr_prop *hints_props(void)
 	    { &hint_add_hosts, "add_hosts", "Load a file with hosts-like formatting and add contents into hints.", },
 	    { &hint_root,   "root", "Replace root hints set (empty value to return current list).", },
 	    { &hint_root_file, "root_file", "Replace root hints set from a zonefile.", },
+	    { &hint_use_nodata, "use_nodata", "Synthesise NODATA if name matches, but type doesn't.", },
 	    { NULL, NULL, NULL }
 	};
 	return prop_list;
