@@ -170,16 +170,16 @@ local function tls_forward_target_authtype(idx, target)
 		return 'pin_sha256'
 	elseif (target.insecure and not (target.ca_file or target.hostname or target.pin_sha256)) then
 		return 'insecure'
-	elseif (target.ca_file and target.hostname and not (target.insecure or target.pin_sha256)) then
-		if not (is_nonempty_string_or_table(target.hostname)
-			and is_nonempty_string_or_table(target.ca_file)) then
+	elseif (target.hostname and not (target.insecure or target.pin_sha256)) then
+		if not (is_nonempty_string_or_table(target.hostname)) then
 			error('TLS_FORWARD target authentication is invalid at position '
-			      .. idx .. '; hostname and ca_file must be string or list of strings')
+			      .. idx .. '; hostname must be string or list of strings')
 		end
+		-- if target.ca_file is empty, system CA will be used
 		return 'cert'
 	else
 		error('TLS_FORWARD authentication options at position ' .. idx
-		      .. ' are invalid; specify one of: pin_sha256 / hostname+ca_file / insecure')
+		      .. ' are invalid; specify one of: pin_sha256 / hostname [+ca_file] / insecure')
 	end
 end
 
@@ -238,7 +238,6 @@ function policy.TLS_FORWARD(target)
 			assert(#pins[sockaddr_lua] > 0)
 			net.tls_client(config.string_addr, pins[sockaddr_lua])
 		elseif config.auth_type == 'cert' then
-			assert(#ca_files[sockaddr_lua] > 0)
 			assert(#hostnames[sockaddr_lua] > 0)
 			net.tls_client(config.string_addr, ca_files[sockaddr_lua], hostnames[sockaddr_lua])
 		else
@@ -334,7 +333,7 @@ local function localhost_reversed(_, req)
 	local is_exact   -- exact dname for localhost
 	local is_apex    -- apex of a locally-served localhost zone
 	local is_nonterm -- empty non-terminal name
-	if ffi.C.knot_dname_is_sub(qry.sname, todname('ip6.arpa.')) then
+	if ffi.C.knot_dname_in_bailiwick(qry.sname, todname('ip6.arpa.')) > 0 then
 		-- exact ::1 query (relying on the calling rule)
 		is_exact = true
 		is_apex = true
@@ -350,7 +349,7 @@ local function localhost_reversed(_, req)
 		else
 			is_exact = false
 			is_apex = false
-			is_nonterm = ffi.C.knot_dname_is_sub(dname_rev4_localhost, qry.sname)
+			is_nonterm = ffi.C.knot_dname_in_bailiwick(dname_rev4_localhost, qry.sname) > 0
 		end
 	end
 
@@ -491,10 +490,18 @@ function policy.DROP(_, _)
 	return kres.FAIL
 end
 
+function policy.REFUSE(_, req)
+	local answer = req.answer
+	answer:rcode(kres.rcode.REFUSED)
+	answer:ad(false)
+	return kres.DONE
+end
+
 function policy.TC(state, req)
 	local answer = req.answer
 	if answer.max_size ~= 65535 then
 		answer:tc(1) -- ^ Only UDP queries
+		answer:ad(false)
 		return kres.DONE
 	else
 		return state
